import typing
from dataclasses import dataclass, field

from xdsl.builder import Builder
from xdsl.context import Context
from xdsl.dialects import arith, builtin, func, llvm, memref, scf, stencil
from xdsl.dialects.arith import ConstantOp
from xdsl.dialects.builtin import (
    DenseArrayBase,
    IndexType,
    IntAttr,
    MemRefType,
    f64,
    i32,
    i64,
)
from xdsl.dialects.experimental.hls import (
    HLSExtractStencilValueOp,
    HLSStreamOp,
    HLSStreamReadOp,
    HLSStreamType,
    HLSStreamWriteOp,
    HLSYieldOp,
    PragmaDataflowOp,
    PragmaPipelineOp,
)
from xdsl.dialects.func import CallOp, FuncOp
from xdsl.dialects.llvm import (
    AllocaOp,
    InsertValueOp,
    LLVMArrayType,
    LLVMPointerType,
    LLVMStructType,
    LoadOp,
    UndefOp,
)
from xdsl.dialects.stencil import (
    AccessOp,
    ApplyOp,
    ExternalLoadOp,
    ExternalStoreOp,
    FieldType,
    ReturnOp,
    StencilBoundsAttr,
    TempType,
)
from xdsl.ir import (
    Attribute,
    Block,
    BlockArgument,
    Operation,
    OpResult,
    Region,
    SSAValue,
)
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.transforms.experimental.convert_stencil_to_ll_mlir import (
    AccessOpToMemRef,
    CastOpToMemRef,
    LoadOpToMemRef,
    StencilToMemRefType,
    TrivialExternalLoadOpCleanup,
    TrivialExternalStoreOpCleanup,
    prepare_apply_body,
)
from xdsl.utils.hints import isa

IN = 0
OUT = 1


# def add_pragma_interface(func_arg: BlockArgument, inout: int, kernel: FuncOp):
#    func_call = None
#    if inout is IN:
#        func_call = Call("IN", func_arg, [])
#    elif inout is OUT:
#        func_call = Call("OUT", func_arg, [])
#
#    kernel.body.block.insert_op_before(typing.cast(Call, func_call), typing.cast(Operation, kernel.body.block.first_op))


def gen_duplicate_loop(
    input_stream: HLSStreamOp,
    duplicate_stream_lst: list[HLSStreamOp],
    n: arith.IndexCastOp,
):
    ii = ConstantOp.from_int_and_width(1, i32)

    @Builder.region([IndexType()])
    def for_body(builder: Builder, args: tuple[BlockArgument, ...]):
        hls_pipeline_op = PragmaPipelineOp(ii)

        builder.insert(hls_pipeline_op)
        hls_read = HLSStreamReadOp(input_stream.results[0])
        builder.insert(hls_read)

        for duplicate_stream in duplicate_stream_lst:
            hls_write = HLSStreamWriteOp(hls_read, duplicate_stream)
            hls_write.attributes["duplicate"] = IntAttr(1)
            builder.insert(hls_write)

        yield_op = scf.YieldOp()
        builder.insert(yield_op)

    lb = ConstantOp.from_int_and_width(0, IndexType())
    ub = n
    step = ConstantOp.from_int_and_width(1, IndexType())

    for_duplicate = scf.ForOp(lb, ub, step, [], for_body)

    return [ii, lb, ub, step, for_duplicate]


@dataclass
class StencilExternalLoadToHLSExternalLoad(RewritePattern):
    module: builtin.ModuleOp
    shift_streams: list[list[HLSStreamOp]]
    out_data_streams: list[HLSStreamOp]
    out_global_mem: list[BlockArgument]
    load_data_declaration: bool = False

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ExternalLoadOp, rewriter: PatternRewriter, /):
        field = op.field

        # Find the llvm.ptr to external memory that genrates the argument to the stencil.external_load. For PSyclone, this is
        # an argument to the parent function. TODO: this might need to be tested and generalised for other codes. Also, we are
        # considering that the function argument will be the second to insertvalue, but we're walking up trhough the second to
        # avoid bumping into arith.constants (see the mlir ssa).
        new_op = field
        func_arg = None

        while not isa(func_arg, BlockArgument):
            assert isinstance(new_op.owner, Operation)
            func_arg = new_op.owner.operands[-1]
            new_op = new_op.owner.operands[0]

        if isa(func_arg.type, LLVMPointerType):
            func_arg_elem_type = func_arg.type.type
        else:
            func_arg_elem_type = func_arg.type

        # add_pragma_interface(func_arg, op.attributes["inout"].data, op.parent_op())
        assert isinstance(op.attributes["inout"], IntAttr)
        inout = op.attributes["inout"].data

        if op.attributes["inout"].data is OUT:
            self.out_global_mem.append(func_arg)

        stencil_type = LLVMStructType.from_type_list(
            [
                LLVMArrayType.from_size_and_type(
                    3,
                    LLVMArrayType.from_size_and_type(
                        3, LLVMArrayType.from_size_and_type(3, f64)
                    ),
                )
            ]
        )

        LLVMStructType.from_type_list([func_arg_elem_type])
        LLVMStructType.from_type_list([stencil_type])

        field_type = field.type
        assert isinstance(field_type, MemRefType)
        shape = field_type.get_shape()

        if len(shape) < 3:
            return

        shape_x = ConstantOp.from_int_and_width(shape[0], i32)
        shape_y = ConstantOp.from_int_and_width(shape[1], i32)
        shape_z = ConstantOp.from_int_and_width(shape[2], i32)

        qualify_apply_op_with_shapes(
            typing.cast(FuncOp, op.parent_op()), shape_x, shape_y, shape_z
        )

        two_int = ConstantOp.from_int_and_width(2, i32)
        shift_shape_x = arith.SubiOp(shape_x, two_int)
        # TODO: generalise this
        data_stream = HLSStreamOp.get(f64)
        stencil_stream = HLSStreamOp.get(stencil_type)

        copy_stencil_stream_lst: list[HLSStreamOp] = []
        # TODO: we are generating 3 copies for now. This is what we need for pw_advection, but should be generalised for codes
        # with a different number of components
        n_components = 0
        for _op in typing.cast(FuncOp, op.parent_op()).body.blocks[0].ops:
            if isinstance(_op, ApplyOp):
                apply_op = _op
                for op_in_apply in apply_op.region.blocks[0].ops:
                    if isinstance(op_in_apply, stencil.ReturnOp):
                        return_op = op_in_apply
                        n_components = len(return_op.arg)

        copy_stencil_stream = HLSStreamOp.get(stencil_type)

        one_int = ConstantOp.from_int_and_width(1, i32)
        four_int = ConstantOp.from_int_and_width(4, i32)
        copy_shift_x = arith.SubiOp(shape_x, four_int)
        copy_shift_y = arith.SubiOp(shape_y, four_int)
        copy_shift_z = arith.SubiOp(shape_z, one_int)
        prod_x_y = arith.MuliOp(copy_shift_x, copy_shift_y)
        copy_n = arith.MuliOp(prod_x_y, copy_shift_z)

        inout = op.attributes["inout"].data

        data_stream.attributes["inout"] = op.attributes["inout"]
        stencil_stream.attributes["inout"] = op.attributes["inout"]
        copy_stencil_stream.attributes["inout"] = op.attributes["inout"]

        # We need to indicate that this is a stencil stream and not a data stream. TODO: make this more elegant
        stencil_stream.attributes["stencil"] = op.attributes["inout"]
        copy_stencil_stream.attributes["stencil"] = op.attributes["inout"]

        for _ in range(n_components):
            copy_stencil_stream_lst.append(copy_stencil_stream)

        threedload_call = CallOp(
            "dummy_load_data", [func_arg, data_stream, shape_x, shape_y, shape_z], []
        )

        @Builder.region
        def load_data_region(builder: Builder):
            yield_op = HLSYieldOp.get()
            builder.insert(threedload_call)
            builder.insert(yield_op)

        load_data_dataflow = PragmaDataflowOp(load_data_region)

        shift_buffer_call = CallOp(
            "shift_buffer",
            [data_stream, stencil_stream, shift_shape_x, shape_y, shape_z],
            [],
        )

        @Builder.region
        def shift_buffer_region(builder: Builder):
            yield_op = HLSYieldOp.get()
            builder.insert(shift_buffer_call)
            builder.insert(yield_op)

        shift_buffer_dataflow = PragmaDataflowOp(shift_buffer_region)

        n_idx = arith.IndexCastOp(copy_n, IndexType())

        duplicate_loop = gen_duplicate_loop(
            stencil_stream, copy_stencil_stream_lst, n_idx
        )

        @Builder.region
        def duplicateStream_region(builder: Builder):
            for dup_op in duplicate_loop:
                builder.insert(dup_op)
            yield_op = HLSYieldOp.get()
            builder.insert(yield_op)

        duplicateStream_dataflow = PragmaDataflowOp(duplicateStream_region)

        ndims = len(field_type.get_shape())
        if inout is IN and ndims == 3:
            rewriter.insert_op_before_matched_op(
                [
                    data_stream,
                    stencil_stream,
                    *copy_stencil_stream_lst,
                    shape_x,
                    shape_y,
                    shape_z,
                    two_int,
                    shift_shape_x,
                    # threedload_call,
                    load_data_dataflow,
                    shift_buffer_dataflow,
                    one_int,
                    four_int,
                    copy_shift_x,
                    copy_shift_y,
                    copy_shift_z,
                    prod_x_y,
                    copy_n,
                    duplicateStream_dataflow,
                ]
            )
            self.shift_streams.append(copy_stencil_stream_lst)
        elif inout is OUT:
            out_data_stream = HLSStreamOp.get(f64)
            out_data_stream.attributes["inout"] = op.attributes["inout"]
            out_data_stream.attributes["data"] = op.attributes["inout"]
            rewriter.insert_op_before_matched_op(
                [
                    out_data_stream,
                ]
            )
            self.out_data_streams.append(out_data_stream)

        if not self.load_data_declaration:
            shift_buffer_func = FuncOp.external(
                "shift_buffer",
                [
                    LLVMPointerType.typed(
                        LLVMStructType.from_type_list([data_stream.elem_type])
                    ),
                    LLVMPointerType.typed(
                        LLVMStructType.from_type_list([stencil_stream.elem_type])
                    ),
                    i32,
                    i32,
                    i32,
                ],
                [],
            )
            self.module.body.block.add_op(shift_buffer_func)

            self.load_data_declaration = True


def qualify_apply_op_with_shapes(
    stencil_func: FuncOp,
    shape_x: arith.ConstantOp,
    shape_y: arith.ConstantOp,
    shape_z: arith.ConstantOp,
):
    block = stencil_func.body.block

    for op in block.ops:
        if isinstance(op, ApplyOp):
            op.attributes["shape_x"] = shape_x.value
            op.attributes["shape_y"] = shape_y.value
            op.attributes["shape_z"] = shape_z.value


def add_read_write_ops(
    indices_stream_to_read: list[int],
    indices_stream_to_write: list[int],
    op: ApplyOp,
    rewriter: PatternRewriter,
    boilerplate: list[Operation],
):
    body_block = op.region.blocks[0]
    return_op = next(o for o in body_block.ops if isinstance(o, ReturnOp))
    stencil_return_vals: list[SSAValue] = [val for val in return_op.arg]

    ConstantOp.from_int_and_width(1, i32)

    stencil_idx = 0
    for arg_index_write in indices_stream_to_write:
        stream_to_write: BlockArgument = op.region.block.args[arg_index_write]
        write_op = HLSStreamWriteOp(stencil_return_vals[stencil_idx], stream_to_write)

        rewriter.insert_op(write_op, InsertPoint.at_end(op.region.block))

    for arg_index_read in indices_stream_to_read:
        stream_to_read = op.region.block.args[arg_index_read]

        read_op = HLSStreamReadOp(stream_to_read)
        read_op.attributes["write_data"] = IntAttr(1)

        rewriter.insert_op(read_op, InsertPoint.at_start(op.region.block))


def transform_apply_into_loop(
    op: ApplyOp, rewriter: PatternRewriter, ndim: int, boilerplate: list[Operation]
):
    dim: int = ndim
    assert dim == 3
    body = prepare_apply_body(op)

    assert isinstance(op.attributes["shape_x"], builtin.IntegerAttr)
    assert isinstance(op.attributes["shape_y"], builtin.IntegerAttr)
    assert isinstance(op.attributes["shape_z"], builtin.IntegerAttr)
    size_x = ConstantOp.from_int_and_width(
        op.attributes["shape_x"].value.data, builtin.IndexType()
    )
    size_y = ConstantOp.from_int_and_width(
        op.attributes["shape_y"].value.data, builtin.IndexType()
    )
    ConstantOp.from_int_and_width(
        op.attributes["shape_z"].value.data, builtin.IndexType()
    )
    one_int = ConstantOp.from_int_and_width(1, i32)
    two = ConstantOp.from_int_and_width(2, builtin.IndexType())
    zero = ConstantOp.from_int_and_width(0, builtin.IndexType())
    one = ConstantOp.from_int_and_width(1, builtin.IndexType())

    size_x_2 = arith.SubiOp(size_x, two)
    size_y_1 = arith.SubiOp(size_y, one)

    lower_x = ConstantOp.from_int_and_width(2, builtin.IndexType())
    lower_y = ConstantOp.from_int_and_width(1, builtin.IndexType())
    lower_z = ConstantOp.from_int_and_width(1, builtin.IndexType())
    upper_x = size_x_2
    upper_y = size_y_1
    upper_z = ConstantOp.from_int_and_width(
        op.attributes["shape_z"].value.data, builtin.IndexType()
    )

    p_remainder = AllocaOp(one_int, i32)

    call_get_number_chunks = CallOp(
        "get_number_chunks", [size_y, p_remainder], [builtin.IndexType()]
    )

    lower_chunks = zero
    upper_chunks = call_get_number_chunks

    lowerBounds = [lower_chunks, lower_x, lower_y, lower_z]
    upperBounds = [upper_chunks, upper_x, upper_y, upper_z]

    # The for loop for the y index receives its trip variable from the get_chunk_size function, since the chunking
    # is happening in the y axis. TODO: this is currently intended for the 3D case. It should be extended to the
    # 1D and 2D cases as well.
    y_for_op: scf.ForOp

    # Pipeline the loop
    ii = ConstantOp.from_int_and_width(1, i32)
    hls_pipeline_op = PragmaPipelineOp(ii)

    # current_region = for_body
    current_block = body
    for_op_lst: list[scf.ForOp] = []
    for i in range(1, dim + 1):
        for_op = scf.ForOp(
            lb=lowerBounds[-i],
            ub=upperBounds[-i],
            step=one,
            iter_args=[],
            body=current_block,
        )
        for_op_lst.append(for_op)
        current_block = Block(
            ops=[for_op, scf.YieldOp()], arg_types=[builtin.IndexType()]
        )

        # if i == 2:
        #    y_for_op = for_op

        if i == 1:
            for_op.body.blocks[0].insert_op_before(
                hls_pipeline_op, typing.cast(Operation, for_op.body.blocks[0].first_op)
            )
            for_op.body.blocks[0].insert_op_before(
                ii, typing.cast(Operation, for_op.body.blocks[0].first_op)
            )

    y_for_op = for_op_lst[1]
    p = scf.ParallelOp(
        lower_bounds=[lowerBounds[0]],
        upper_bounds=[upperBounds[0]],
        steps=[one],
        body=Region(current_block),
    )

    chunk_num = p.body.block.args[0]

    MAX_Y_SIZE = 16
    max_chunk_length = ConstantOp.from_int_and_width(MAX_Y_SIZE, i32)

    remainder = LoadOp(p_remainder)

    call_get_chunk_size = CallOp(
        "get_chunk_size",
        [chunk_num, call_get_number_chunks, max_chunk_length, remainder],
        [builtin.IndexType()],
    )

    p.body.block.insert_op_before(
        call_get_chunk_size, typing.cast(Operation, p.body.block.first_op)
    )
    chunk_size_y_1 = arith.SubiOp(call_get_chunk_size, one)
    p.body.block.insert_op_after(chunk_size_y_1, call_get_chunk_size)

    old_operands_lst = [old_operand for old_operand in y_for_op.operands]
    y_for_op.operands = (
        [old_operands_lst[0]] + [chunk_size_y_1.results[0]] + old_operands_lst[2:]
    )

    p.attributes["compute_loop"] = IntAttr(1)

    @Builder.region
    def p_region(builder: Builder):
        builder.insert(p)
        builder.insert(HLSYieldOp.get())

    p_dataflow = PragmaDataflowOp(p_region)

    boilerplate += [
        size_x,
        size_y,
        one_int,
        one,
        two,
        max_chunk_length,
        p_remainder,
        upper_x,
        *lowerBounds,
        upper_chunks,
        upper_y,
        upper_z,
        remainder,
    ]

    return p_dataflow


@dataclass
class ApplyOpToHLS(RewritePattern):
    module: builtin.ModuleOp
    shift_streams: list[list[HLSStreamOp]]
    out_data_streams: list[HLSStreamOp]
    out_global_mem: list[BlockArgument]

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ApplyOp, rewriter: PatternRewriter, /):
        # We qualify the parent function as a kernel for futher processing
        typing.cast(Operation, op.parent_op()).attributes["kernel"] = IntAttr(1)

        body_block = op.region.blocks[0]
        return_op = next(o for o in body_block.ops if isinstance(o, ReturnOp))
        n_components = len(return_op.arg)
        apply_clones_lst: list[ApplyOp] = [op.clone() for _ in range(n_components)]

        # We replace the temp arguments by HLS streams. Only for the 3D temps
        for k in range(len(apply_clones_lst)):
            # Insert the HLS stream operands and their corresponding block arguments for
            # reading from the shift buffer and writing to external memory.
            # We replace by streams only the 3D temps.
            # The rest should be left as is operand_stream = dict()
            current_stream = 0

            new_operands_lst: list[OpResult] = []
            apply_clone = apply_clones_lst[k]

            for i in range(len(apply_clone.operands)):
                operand: OpResult = typing.cast(OpResult, apply_clone.operands[i])
                operand_type = operand.type
                assert isinstance(operand_type, TempType)
                assert isinstance(operand_type.bounds, StencilBoundsAttr)
                n_dims = len(operand_type.bounds.lb)

                if n_dims == 3:
                    stream = self.shift_streams[current_stream][k]
                    rewriter.modify_value_type(
                        apply_clone.region.block.args[i], stream.results[0].type
                    )

                    assert isa(stream.results, list[OpResult])
                    new_operands_lst.append(stream.results[0])
                    current_stream += 1
                else:
                    new_operands_lst.append(operand)

            apply_clone.operands = new_operands_lst + [
                self.out_data_streams[k].results[0]
            ]

        indices_stream_to_read: list[int] = []
        indices_stream_to_write: list[int] = []
        i = 0
        apply_clone = apply_clones_lst[-1]
        for _operand in apply_clone.operands:
            assert isinstance(_operand, BlockArgument) or isinstance(_operand, OpResult)
            if isinstance(_operand, OpResult) and isinstance(_operand.op, HLSStreamOp):
                assert isinstance(_operand.op.attributes["inout"], IntAttr)
                if (
                    "stencil" in _operand.op.attributes
                    and _operand.op.attributes["inout"].data is IN
                ):
                    indices_stream_to_read.append(i)
                if (
                    "data" in _operand.op.attributes
                    and _operand.op.attributes["inout"].data is OUT
                ):
                    indices_stream_to_write.append(i)
            i += 1

        for write_idx in indices_stream_to_write:
            apply_clone.region.blocks[0].insert_arg(
                self.out_data_streams[0].results[0].type, write_idx
            )

        # Get this apply's ReturnOp
        body_block = op.region.blocks[0]
        return_op = next(o for o in body_block.ops if isinstance(o, ReturnOp))

        # We are going to split the apply by the operations conducive to each returned value
        boilerplate: list[list[Operation]] = [[] for _ in range(len(return_op.arg))]
        new_apply_lst: list[ApplyOp] = []
        new_return_component_lst: list[ReturnOp] = []

        component_idx = 0
        k = 0
        new_apply: ApplyOp
        for component in return_op.arg:
            assert isinstance(component, OpResult)
            new_apply = apply_clones_lst[k]
            k += 1

            component_operations: dict[int, Operation] = dict()
            operation_indices: set[int] = set()
            component_operations[
                typing.cast(Block, return_op.parent_block()).get_operation_index(
                    return_op
                )
            ] = return_op
            operation_indices.add(
                typing.cast(Block, return_op.parent_block()).get_operation_index(
                    return_op
                )
            )
            collectComponentOperations(
                component, component_operations, operation_indices
            )

            new_apply_block = new_apply.region.blocks[0]
            new_return_op = next(
                o for o in new_apply_block.ops if isinstance(o, ReturnOp)
            )

            new_return_op.detach()
            new_return_op.erase()

            for operation in reversed(new_apply_block.ops):
                op_index = new_apply_block.get_operation_index(operation)
                if op_index not in operation_indices:
                    operation.detach()
                    operation.erase()

            new_component = typing.cast(Operation, new_apply_block.last_op).results[0]
            new_return_component = ReturnOp.get([new_component])
            new_apply_block.add_op(new_return_component)

            add_read_write_ops(
                indices_stream_to_read,
                indices_stream_to_write,
                new_apply,
                rewriter,
                boilerplate[component_idx],
            )
            new_apply_lst.append(new_apply)
            new_return_component_lst.append(new_return_component)
            component_idx += 1

        get_number_chunks = FuncOp.external(
            "get_number_chunks",
            [builtin.IndexType(), LLVMPointerType.typed(i32)],
            [builtin.IndexType()],
        )

        get_chunk_size = FuncOp.external(
            "get_chunk_size",
            [builtin.IndexType(), builtin.IndexType(), i32, i32],
            [builtin.IndexType()],
        )

        self.module.body.block.add_op(get_number_chunks)
        self.module.body.block.add_op(get_chunk_size)

        ndims: int = typing.cast(TempType[Attribute], op.res[0].type).get_num_dims()

        rewriter.erase_op(return_op)
        for new_return_component in new_return_component_lst:
            rewriter.erase_op(new_return_component)

        p_dataflow_lst: list[PragmaDataflowOp] = []
        component_idx = 0
        for new_apply in new_apply_lst:
            p_dataflow = transform_apply_into_loop(
                new_apply, rewriter, ndims, boilerplate[component_idx]
            )
            p_dataflow_lst.append(p_dataflow)
            component_idx += 1

        operations_to_insert: list[Operation] = []
        for i in range(n_components):
            operations_to_insert += boilerplate[i] + [p_dataflow_lst[i]]

        rewriter.insert_op_before_matched_op(operations_to_insert)

        rewriter.replace_matched_op(new_apply_lst[-1])


def collectComponentOperations(
    op: OpResult,
    component_operations: dict[int, Operation],
    operation_indices: set[int],
):
    parent_op = op.owner

    for operand in parent_op.operands:
        if not isinstance(operand, BlockArgument):
            assert isinstance(operand, OpResult)
            block_index = typing.cast(
                Block, operand.op.parent_block()
            ).get_operation_index(operand.op)
            component_operations[block_index] = operand.op
            operation_indices.add(block_index)
            collectComponentOperations(operand, component_operations, operation_indices)


def get_number_external_stores(op: FuncOp):
    external_stores_lst = [
        o for o in op.body.blocks[0].ops if isinstance(o, ExternalStoreOp)
    ]

    return len(external_stores_lst)


@dataclass
class StencilExternalStoreToHLSWriteData(RewritePattern):
    module: builtin.ModuleOp
    out_data_streams: list[HLSStreamType]
    write_data_declaration: bool = False
    func_args_lst: list[BlockArgument] = field(default_factory=list)
    n_args: int = 0
    total_args: int | None = None

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ExternalStoreOp, rewriter: PatternRewriter, /):
        if self.total_args is None:
            self.total_args = get_number_external_stores(
                typing.cast(func.FuncOp, op.parent_op())
            )
        temp = op.temp

        # Find the llvm.ptr to external memory that genrates the argument to the stencil.external_load. For PSyclone, this is
        # an argument to the parent function. TODO: this might need to be tested and generalised for other codes. Also, we are
        # considering that the function argument will be the second to insertvalue, but we're walking up trhough the second to
        # avoid bumping into arith.constants (see the mlir ssa).
        new_op = temp
        func_arg = None

        while not isa(func_arg, BlockArgument):
            assert isinstance(new_op.owner, Operation)
            func_arg = new_op.owner.operands[-1]
            new_op = new_op.owner.operands[0]

        self.func_args_lst.append(func_arg)
        self.n_args += 1

        if self.n_args == self.total_args:
            write_data_func_name = f"write_data_{self.total_args}"
            packed_type = LLVMPointerType.typed(
                LLVMStructType.from_type_list(
                    [LLVMArrayType.from_size_and_type(8, f64)]
                )
            )
            assert isinstance(self.out_data_streams[0], HLSStreamOp)
            out_data_type = self.out_data_streams[0].elem_type
            LLVMPointerType.typed(out_data_type)
            out_data_stream_type = LLVMPointerType.typed(
                LLVMStructType.from_type_list([out_data_type])
            )
            write_data_func_args_lst = (
                self.total_args * [out_data_stream_type]
                + self.total_args * [packed_type]
                + 3 * [i32]
            )
            write_data_func = FuncOp.external(
                write_data_func_name,
                write_data_func_args_lst,
                [],
            )
            self.module.body.block.add_op(write_data_func)

            func_arg = None

            while not isa(func_arg, BlockArgument):
                assert isinstance(new_op.owner, Operation)
                func_arg = new_op.owner.operands[-1]

            op_field_type = op.field.type
            assert isinstance(op_field_type, MemRefType)
            shape = op_field_type.shape
            shape_x = ConstantOp.from_int_and_width(shape.data[0].data, i32)
            shape_y = ConstantOp.from_int_and_width(shape.data[1].data, i32)
            shape_z = ConstantOp.from_int_and_width(shape.data[2].data, i32)

            call_write_data = CallOp(
                write_data_func_name,
                [
                    *[
                        typing.cast(Operation, stream).results[0]
                        for stream in self.out_data_streams
                    ],
                    *reversed(self.func_args_lst),
                    shape_x,
                    shape_y,
                    shape_z,
                ],
                [],
            )

            @Builder.region
            def write_data_df_region(builder: Builder):
                builder.insert(call_write_data)
                hls_yield_op = HLSYieldOp.get()
                builder.insert(hls_yield_op)

            write_data_dataflow = PragmaDataflowOp(write_data_df_region)

            rewriter.insert_op_after_matched_op(
                [shape_x, shape_y, shape_z, write_data_dataflow]
            )


@dataclass
class StencilAccessOpToReadBlockOp(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: AccessOp, rewriter: PatternRewriter, /):
        result_hls_read: OpResult | None = None

        replace_access = False

        for use in op.temp.uses:
            if (
                isinstance(use.operation, HLSStreamReadOp)
                and use.operation.parent_op() == op.parent_op()
            ):
                hls_read = use.operation

                assert isinstance(hls_read.results[0], OpResult)
                result_hls_read = hls_read.results[0]
                replace_access = True

        if replace_access:
            access_idx: list[int] = []
            assert isa(op.offset.array, builtin.ArrayAttr[IntAttr])
            for idx in op.offset.array.data:
                access_idx.append(idx.data + 1)

            access_idx_array = DenseArrayBase.create_dense_int(i64, [0] + access_idx)

            assert isinstance(result_hls_read, OpResult)
            stencil_value = HLSExtractStencilValueOp(
                access_idx_array, result_hls_read, f64
            )

            rewriter.replace_matched_op(stencil_value)


# Copied from convert_stencil_to_ll_mlir
@dataclass
class StencilStoreToSubview(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: FuncOp, rewriter: PatternRewriter, /):
        stores = [o for o in op.walk() if isinstance(o, stencil.StoreOp)]

        for store in stores:
            field = store.field
            assert isa(field.type, FieldType[Attribute])
            assert isa(field.type.bounds, StencilBoundsAttr)
            temp = store.temp
            assert isa(temp.type, TempType[Attribute])
            offsets = [i for i in -field.type.bounds.lb]
            sizes = [i for i in temp.type.get_shape()]
            subview = memref.SubviewOp.from_static_parameters(
                field,
                StencilToMemRefType(field.type),
                offsets,
                sizes,
                [1] * len(sizes),
            )
            name = None
            if subview.source.name_hint:
                name = subview.source.name_hint + "_storeview"
            subview.result.name_hint = name
            if isinstance(field.owner, Operation):
                rewriter.insert_op(subview, InsertPoint.after(field.owner))
            else:
                rewriter.insert_op(subview, InsertPoint.at_start(field.owner))

            rewriter.erase_op(store)


@dataclass
class TrivialStoreOpCleanup(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: stencil.StoreOp, rewriter: PatternRewriter, /):
        rewriter.erase_matched_op()


@dataclass
class TrivialApplyOpCleanup(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ApplyOp, rewriter: PatternRewriter, /):
        rewriter.erase_matched_op()


@dataclass
class QualifyAllArgumentsAsOut(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ExternalLoadOp, rewriter: PatternRewriter, /):
        op.attributes["inout"] = IntAttr(IN)


@dataclass
class GetInoutAttributeFromExternalStore(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ExternalStoreOp, rewriter: PatternRewriter, /):
        for use in op.field.uses:
            if isinstance(use.operation, ExternalLoadOp):
                use.operation.attributes["inout"] = IntAttr(OUT)


def get_number_input_stencils(op: FuncOp):
    # ndims = len(field.typ.get_shape())
    def dim(o: ExternalLoadOp):
        assert isinstance(o.field, OpResult)
        op_field_type = o.field.type
        assert isinstance(op_field_type, memref.MemRefType)
        return len(op_field_type.get_shape())

    external_load_lst = [
        o
        for o in op.body.blocks[0].ops
        if isinstance(o, ExternalLoadOp) and dim(o) == 3
    ]

    n = sum(
        [
            1
            for o in external_load_lst
            if typing.cast(IntAttr, o.attributes["inout"]).data == IN
        ]
    )

    return n


@dataclass
class GroupLoadsUnderSameDataflow(RewritePattern):
    module: builtin.ModuleOp
    first_load: CallOp | None = None
    # sizes: IntegerType | None = None
    sizes: list[OpResult | SSAValue] = field(default_factory=list)
    in_module_load_all_data_func: FuncOp | None = None
    n_current_load: int = 0
    n_input: int = -1

    data_arrays: list[BlockArgument] = field(default_factory=list)
    data_streams: list[OpResult] = field(default_factory=list)

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: CallOp, rewriter: PatternRewriter, /):
        if op.callee.root_reference.data == "dummy_load_data":
            self.n_input = get_number_input_stencils(
                typing.cast(
                    func.FuncOp,
                    typing.cast(PragmaDataflowOp, op.parent_op()).parent_op(),
                )
            )
        if (
            op.callee.root_reference.data == "dummy_load_data"
            and self.n_current_load < self.n_input
        ):
            self.n_current_load += 1

            self.data_arrays.append(typing.cast(BlockArgument, op.operands[0]))
            self.data_streams.append(typing.cast(OpResult, op.operands[1]))

            # We are using the same sizes for all the load_data operations. We remove the duplicates
            if self.first_load is None:
                self.first_load = op
                self.sizes += list(
                    op.operands[2:]
                )  # [operand for operand in op.operands[2:]]
            else:
                parent_dataflow = typing.cast(PragmaDataflowOp, op.parent_op())
                rewriter.erase_matched_op()
                parent_dataflow.detach()
                parent_dataflow.erase()

            if self.n_current_load == self.n_input:
                assert isinstance(self.first_load, CallOp)
                assert isinstance(self.first_load.arguments[1].type, HLSStreamType)
                load_data_func_name = f"load_data_{self.n_input}"
                hls_stream_type = LLVMPointerType.typed(
                    LLVMStructType.from_type_list(
                        [self.first_load.arguments[1].type.element_type]
                    )
                )
                load_data_args_lst = (
                    self.n_input * [self.first_load.arguments[0].type]
                    + self.n_input * [hls_stream_type]
                    + 3 * [i32]
                )
                load_data_func = FuncOp.external(
                    load_data_func_name,
                    load_data_args_lst,
                    [],
                )

                self.module.body.block.add_op(load_data_func)

                load_all_data_args = self.data_arrays + self.data_streams + self.sizes
                assert isa(load_all_data_args, list[SSAValue])

                call_load_all_data = CallOp(
                    load_data_func_name,
                    load_all_data_args,
                    [],
                )

                parent_dataflow = typing.cast(
                    PragmaDataflowOp, self.first_load.parent_op()
                )

                rewriter.replace_op(self.first_load, call_load_all_data)

                for data_stream in self.data_streams:
                    data_stream.op.detach()
                    rewriter.insert_op(
                        data_stream.op, InsertPoint.before(parent_dataflow)
                    )


@dataclass
class PackData(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: ExternalLoadOp, rewriter: PatternRewriter, /):
        field = op.field

        # Find the llvm.ptr to external memory that genrates the argument to the stencil.external_load. For PSyclone, this is
        # an argument to the parent function. TODO: this might need to be tested and generalised for other codes. Also, we are
        # considering that the function argument will be the second to insertvalue, but we're walking up trhough the second to
        # avoid bumping into arith.constants (see the mlir ssa).
        new_op = field
        func_arg = None

        while not isa(func_arg, BlockArgument):
            assert isinstance(new_op.owner, Operation)
            func_arg = new_op.owner.operands[-1]
            new_op = new_op.owner.operands[0]

        field_type = field.type
        assert isinstance(field_type, memref.MemRefType)
        shape = field_type.get_shape()
        ndims = len(shape)

        arg_idx = func_arg.index
        parent_func = typing.cast(func.FuncOp, op.parent_op())

        assert isa(op.attributes["inout"], IntAttr)
        if op.attributes["inout"].data == OUT or (
            op.attributes["inout"].data == IN and ndims == 3
        ):
            # TODO: this should be generalised by packaging the original type instead of
            # f64. We would need intrinsics to deal with the different types
            packed_type = LLVMPointerType.typed(
                LLVMStructType.from_type_list(
                    [LLVMArrayType.from_size_and_type(8, f64)]
                )
            )
            parent_func.replace_argument_type(arg_idx, packed_type)

            for use in func_arg.uses:
                if isinstance(use.operation, InsertValueOp):
                    insertvalue = use.operation

                    assert isinstance(insertvalue.container, OpResult)
                    container_op = insertvalue.container.op
                    if isinstance(container_op, UndefOp):
                        # We mark the UndefOp to update its type in the next pass and
                        # also update the type returned by the insertvalue operation
                        # that uses it

                        container_op.attributes["replace"] = IntAttr(0)

                        # Update the return types of the chain of insertvalues used to generatate the field structure
                        assert isinstance(insertvalue.res.type, llvm.LLVMStructType)
                        field_struct = insertvalue.res.type

                        old_type = list(field_struct.types.data)
                        new_type = old_type
                        new_type[0] = packed_type
                        new_type[1] = packed_type
                        struct_new_type = LLVMStructType.from_type_list(new_type)

                        current_insertvalue = insertvalue
                        current_insertvalue.res.type = struct_new_type
                        update_types_insertvalue(current_insertvalue, struct_new_type)


def update_types_insertvalue(op: InsertValueOp, new_type: llvm.LLVMStructType):
    for use in op.res.uses:
        if isinstance(use.operation, InsertValueOp):
            use.operation.res.type = new_type
            update_types_insertvalue(use.operation, new_type)


@dataclass
class PackDataInStencilField(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: UndefOp, rewriter: PatternRewriter, /):
        if "replace" in op.attributes:
            # packed_type = LLVMPointerType.typed(LLVMArrayType.from_size_and_type(8, f64))
            packed_type = LLVMPointerType.typed(
                LLVMStructType.from_type_list(
                    [LLVMArrayType.from_size_and_type(8, f64)]
                )
            )
            assert isinstance(op.res.type, llvm.LLVMStructType)
            field_struct = op.res.type

            old_type = list(field_struct.types.data)
            new_type = old_type
            new_type[0] = packed_type
            new_type[1] = packed_type

            struct_new_type = LLVMStructType.from_type_list(new_type)

            new_container_op = UndefOp(struct_new_type)

            rewriter.replace_matched_op(new_container_op)


# We create copies for all the coefficients. We create more than one copy where necesssary
@dataclass
class GetRepeatedCoefficients(RewritePattern):
    original_memref_lst: list[memref.CastOp]
    clone_memref_lst: list[memref.AllocaOp]

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: memref.SubviewOp, rewriter: PatternRewriter, /):
        assert isinstance(op.source, OpResult)
        assert isinstance(op.source.op, memref.CastOp)
        cast = op.source.op  # original memref
        cast_dest_type = cast.dest.type
        assert isinstance(cast_dest_type, memref.MemRefType)
        cast.clone()

        dim = len(cast_dest_type.shape.data)
        cast_dest_type.shape.data[0].data
        if dim == 1:
            uses_copy = set(op.results[0].uses)
            for use in uses_copy:
                if isinstance(use.operation, memref.LoadOp):
                    memref_copy = memref.AllocaOp.get(
                        return_type=f64, shape=cast_dest_type.shape
                    )
                    use.operation.operands[0] = memref_copy.results[0]
                    rewriter.insert_op_before_matched_op(memref_copy)

                    self.original_memref_lst.append(cast)
                    self.clone_memref_lst.append(memref_copy)
                elif isinstance(use.operation, stencil.ApplyOp):
                    op.results[0].remove_use(use)

            rewriter.erase_matched_op()


@dataclass
class MakeLocaCopiesOfCoefficients(RewritePattern):
    original_memref_lst: list[memref.CastOp]
    clone_memref_lst: list[memref.AllocaOp]
    inserted_already = False

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: PragmaDataflowOp, rewriter: PatternRewriter, /):
        if (
            "compute_loop"
            in typing.cast(Operation, op.body.blocks[0].first_op).attributes
            and not self.inserted_already
            and len(self.original_memref_lst) > 0
        ):
            original_memref_lst_dest_type = self.original_memref_lst[0].dest.type
            assert isinstance(original_memref_lst_dest_type, memref.MemRefType)
            dim = original_memref_lst_dest_type.shape.data[0].data

            lb = ConstantOp.from_int_and_width(0, IndexType())
            ub = ConstantOp.from_int_and_width(dim, IndexType())
            step = ConstantOp.from_int_and_width(1, IndexType())

            ii = ConstantOp.from_int_and_width(1, i32)

            @Builder.region([IndexType()])
            def for_body(builder: Builder, args: tuple[BlockArgument, ...]):
                hls_pipeline_op = PragmaPipelineOp(ii)
                builder.insert(hls_pipeline_op)
                for i in range(len(self.original_memref_lst)):
                    load_op = memref.LoadOp.get(self.original_memref_lst[i], [args[0]])
                    store_op = memref.StoreOp.get(
                        load_op, self.clone_memref_lst[i], [args[0]]
                    )
                    builder.insert(load_op)
                    builder.insert(store_op)

                yield_op = scf.YieldOp()
                builder.insert(yield_op)

            for_local_copies = scf.ForOp(lb, ub, step, [], for_body)
            rewriter.insert_op_before_matched_op([lb, ub, step, ii, for_local_copies])

            self.inserted_already = True


@dataclass
class TrivialCleanUpAuxAttributes(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: scf.ParallelOp, rewriter: PatternRewriter, /):
        if "compute_loop" in op.attributes:
            del op.attributes["compute_loop"]


@dataclass
class QualifyInterfacesPass(RewritePattern):
    module: builtin.ModuleOp
    declared_coeff_func: bool = False
    called_coeff_func: bool = False
    interface_coeff_func_name = "_maxi_coeff"

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: func.FuncOp, rewriter: PatternRewriter, /):
        bundle_idx = 1
        arg_idx = 0

        if "kernel" in op.attributes:
            del op.attributes["kernel"]

            for input_arg in op.function_type.inputs:
                if isinstance(input_arg, LLVMPointerType) and isinstance(
                    input_arg.type, LLVMStructType
                ):
                    interface_func_name = f"_maxi_gmem{bundle_idx}"
                    interface_func = func.FuncOp.external(
                        interface_func_name, [input_arg], []
                    )
                    self.module.body.block.add_op(interface_func)

                    call_interface_func = func.CallOp(
                        interface_func_name, [op.body.blocks[0].args[arg_idx]], []
                    )
                    rewriter.insert_op(
                        call_interface_func, InsertPoint.at_start(op.body.blocks[0])
                    )

                    bundle_idx += 1
                else:
                    call_interface_func = llvm.CallOp(
                        self.interface_coeff_func_name, op.body.blocks[0].args[arg_idx]
                    )
                    rewriter.insert_op(
                        call_interface_func, InsertPoint.at_start(op.body.blocks[0])
                    )
                    self.called_coeff_func = True

                arg_idx += 1

        if self.called_coeff_func and not self.declared_coeff_func:
            interface_coeff_func_type = llvm.LLVMFunctionType([], None, True)
            interface_coeff_func = llvm.FuncOp(
                self.interface_coeff_func_name,
                interface_coeff_func_type,
                llvm.LinkageAttr("external"),
            )
            self.module.body.block.add_op(interface_coeff_func)
            self.declared_coeff_func = True


@dataclass(frozen=True)
class HLSConvertStencilToLLMLIRPass(ModulePass):
    name = "hls-convert-stencil-to-ll-mlir"

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        module: builtin.ModuleOp = op
        shift_streams: list[list[HLSStreamOp]] = []
        out_data_streams: list[HLSStreamOp] = []
        out_global_mem: list[BlockArgument] = []

        inout_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    QualifyAllArgumentsAsOut(),
                    GetInoutAttributeFromExternalStore(),
                ]
            ),
            apply_recursively=False,
            walk_reverse=False,
        )
        inout_pass.rewrite_module(op)

        pack_data_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    PackData(),
                    PackDataInStencilField(),
                ]
            ),
            apply_recursively=True,
            walk_reverse=True,
        )
        pack_data_pass.rewrite_module(op)

        hls_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    StencilExternalLoadToHLSExternalLoad(
                        module, shift_streams, out_data_streams, out_global_mem
                    ),
                ]
            ),
            apply_recursively=False,
            walk_reverse=True,
        )
        hls_pass.rewrite_module(op)

        adapt_stencil_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    ApplyOpToHLS(
                        module, shift_streams, out_data_streams, out_global_mem
                    ),
                    StencilAccessOpToReadBlockOp(),
                    StencilStoreToSubview(),
                    CastOpToMemRef(),
                    LoadOpToMemRef(),
                    AccessOpToMemRef(),
                ]
            ),
            apply_recursively=False,
            walk_reverse=True,
        )
        adapt_stencil_pass.rewrite_module(op)

        write_data_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    StencilExternalStoreToHLSWriteData(module, []),
                    # TrivialExternalLoadOpCleanup(),
                    TrivialExternalStoreOpCleanup(),
                    TrivialStoreOpCleanup(),
                ]
            ),
            apply_recursively=False,
            walk_reverse=True,
        )
        write_data_pass.rewrite_module(op)

        grouploads_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier([GroupLoadsUnderSameDataflow(op)]),
            apply_recursively=False,
            walk_reverse=False,
        )
        grouploads_pass.rewrite_module(op)

        cleanup_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    TrivialExternalLoadOpCleanup(),
                    TrivialExternalStoreOpCleanup(),
                    TrivialStoreOpCleanup(),
                ]
            ),
            apply_recursively=False,
            walk_reverse=False,
        )
        cleanup_pass.rewrite_module(op)

        clean_apply_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [TrivialApplyOpCleanup()]  # , GroupLoadsUnderSameDataflow(op)]
            ),
            apply_recursively=False,
            walk_reverse=False,
        )
        clean_apply_pass.rewrite_module(op)

        original_memref_lst: list[memref.CastOp] = []
        clone_memref_lst: list[memref.AllocaOp] = []

        get_repeated = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    GetRepeatedCoefficients(original_memref_lst, clone_memref_lst),
                ]
            ),
            apply_recursively=True,
            walk_reverse=True,
        )
        get_repeated.rewrite_module(op)

        make_local_copies = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [MakeLocaCopiesOfCoefficients(original_memref_lst, clone_memref_lst)]
            ),
            apply_recursively=True,
            walk_reverse=False,
        )
        make_local_copies.rewrite_module(op)

        final_cleanup = PatternRewriteWalker(
            GreedyRewritePatternApplier([TrivialCleanUpAuxAttributes()]),
            apply_recursively=True,
            walk_reverse=False,
        )
        final_cleanup.rewrite_module(op)

        interfaces_pass = PatternRewriteWalker(
            QualifyInterfacesPass(op),
            apply_recursively=True,
            walk_reverse=False,
        )
        interfaces_pass.rewrite_module(op)



from collections.abc import Sequence

from xdsl.dialects.builtin import ModuleOp
from xdsl.interactive.passes import AvailablePass
from xdsl.pattern_rewriter import PatternRewriter, RewritePattern
from xdsl.traits import HasCanonicalizationPatternsTrait
from xdsl.transforms import individual_rewrite


def get_all_possible_rewrites(
    module: ModuleOp,
    rewrite_by_name: dict[str, dict[str, RewritePattern]],
) -> Sequence[AvailablePass]:
    """
    Function that takes a sequence of IndividualRewrite Patterns and a ModuleOp, and
    returns the possible rewrites.
    Issue filed: https://github.com/xdslproject/xdsl/issues/2162
    """

    res: list[AvailablePass] = []

    for op_idx, matched_op in enumerate(module.walk()):
        pattern_by_name = rewrite_by_name.get(matched_op.name, {}).copy()

        if (
            trait := matched_op.get_trait(HasCanonicalizationPatternsTrait)
        ) is not None:
            for pattern in trait.get_canonicalization_patterns():
                pattern_by_name[type(pattern).__name__] = pattern

        for pattern_name, pattern in pattern_by_name.items():
            cloned_op = tuple(module.clone().walk())[op_idx]
            rewriter = PatternRewriter(cloned_op)
            pattern.match_and_rewrite(cloned_op, rewriter)
            if rewriter.has_done_action:
                res.append(
                    AvailablePass(
                        f"{cloned_op}:{cloned_op.name}:{pattern_name}",
                        individual_rewrite.ApplyIndividualRewritePass(
                            op_idx, cloned_op.name, pattern_name
                        ),
                    )
                )

    return res



from dataclasses import dataclass

from xdsl.backend.riscv.register_allocation import RegisterAllocatorLivenessBlockNaive
from xdsl.backend.riscv.riscv_register_queue import RiscvRegisterQueue
from xdsl.context import Context
from xdsl.dialects import riscv_func
from xdsl.dialects.builtin import ModuleOp
from xdsl.passes import ModulePass


@dataclass(frozen=True)
class RISCVRegisterAllocation(ModulePass):
    """
    Allocates unallocated registers in the module.
    """

    name = "riscv-allocate-registers"

    allocation_strategy: str = "LivenessBlockNaive"

    limit_registers: int | None = None

    exclude_preallocated: bool = True
    """
    Enables tracking of already allocated registers and excludes them from the
    available set.
    This does not keep track of any liveness information and the preallocated registers
    are excluded completely from any further allocation decisions.
    """

    exclude_snitch_reserved: bool = True
    """Excludes floating-point registers that are used by the Snitch ISA extensions."""

    add_regalloc_stats: bool = False
    """
    Inserts a comment with register allocation info in the IR.
    """

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        allocator_strategies = {
            "LivenessBlockNaive": RegisterAllocatorLivenessBlockNaive,
        }

        if self.allocation_strategy not in allocator_strategies:
            raise ValueError(
                f"Unknown register allocation strategy {self.allocation_strategy}. "
                f"Available allocation types: {allocator_strategies.keys()}"
            )

        if self.limit_registers is not None and self.limit_registers < 0:
            raise ValueError(
                "The limit of available registers cannot be less than 0."
                "When set to 0 it signifies all available registers are used."
            )

        for inner_op in op.walk():
            if isinstance(inner_op, riscv_func.FuncOp):
                riscv_register_queue = RiscvRegisterQueue.default()
                if self.limit_registers is not None:
                    riscv_register_queue.limit_registers(self.limit_registers)
                allocator = allocator_strategies[self.allocation_strategy](
                    riscv_register_queue
                )
                allocator.exclude_preallocated = self.exclude_preallocated
                allocator.exclude_snitch_reserved = self.exclude_snitch_reserved
                allocator.allocate_func(
                    inner_op, add_regalloc_stats=self.add_regalloc_stats
                )



import pytest

from xdsl.builder import ImplicitBuilder
from xdsl.dialects import arith, memref_stream
from xdsl.dialects.builtin import (
    AffineMapAttr,
    ArrayAttr,
    Float32Type,
    IndexType,
    IntAttr,
    IntegerAttr,
    MemRefType,
    ModuleOp,
    i32,
)
from xdsl.interpreter import Interpreter
from xdsl.interpreters.arith import ArithFunctions
from xdsl.interpreters.memref_stream import MemRefStreamFunctions
from xdsl.interpreters.shaped_array import ShapedArray
from xdsl.interpreters.utils.ptr import TypedPtr
from xdsl.ir import Block, Region
from xdsl.ir.affine import AffineExpr, AffineMap
from xdsl.utils.test_value import TestSSAValue

indextype = IndexType()


def index(value: int) -> IntegerAttr[IndexType]:
    return IntegerAttr(value, indextype)


def test_memref_stream_generic():
    interpreter = Interpreter(ModuleOp([]))
    interpreter.register_implementations(MemRefStreamFunctions())
    interpreter.register_implementations(ArithFunctions())

    op = memref_stream.GenericOp(
        (
            TestSSAValue(MemRefType(i32, [2, 3])),
            TestSSAValue(MemRefType(i32, [3, 2])),
        ),
        (TestSSAValue(MemRefType(i32, [1, 6])),),
        (),
        Region(Block(arg_types=(i32, i32, i32))),
        ArrayAttr(
            (
                AffineMapAttr(AffineMap.identity(2)),
                AffineMapAttr(AffineMap.transpose_map()),
                AffineMapAttr(
                    AffineMap(
                        2,
                        0,
                        (
                            AffineExpr.constant(0),
                            AffineExpr.dimension(0) * 3 + AffineExpr.dimension(1),
                        ),
                    )
                ),
            )
        ),
        ArrayAttr(
            (
                memref_stream.IteratorTypeAttr.parallel(),
                memref_stream.IteratorTypeAttr.parallel(),
            )
        ),
        ArrayAttr((index(2), index(3))),
        ArrayAttr(()),
    )

    with ImplicitBuilder(op.body) as (a, b, _c_init):
        c = arith.MuliOp(a, b).result
        memref_stream.YieldOp(c)

    op.verify()

    a = ShapedArray(TypedPtr.new_float64([1, 2, 3, 4, 5, 6]), [2, 3])
    b = ShapedArray(TypedPtr.new_float64([1, 4, 2, 5, 3, 6]), [3, 2])
    c = ShapedArray(TypedPtr.new_float64([-1, -1, -1, -1, -1, -1]), [1, 6])

    interpreter.run_op(op, (a, b, c))

    assert c.data == [1, 4, 9, 16, 25, 36]


def test_memref_stream_generic_scalar():
    interpreter = Interpreter(ModuleOp([]))
    interpreter.register_implementations(MemRefStreamFunctions())
    interpreter.register_implementations(ArithFunctions())

    op = memref_stream.GenericOp(
        (
            TestSSAValue(MemRefType(i32, [2, 3])),
            TestSSAValue(i32),
        ),
        (TestSSAValue(MemRefType(i32, [1, 6])),),
        (),
        Region(Block(arg_types=(i32, i32, i32))),
        ArrayAttr(
            (
                AffineMapAttr(AffineMap.identity(2)),
                AffineMapAttr(AffineMap.from_callable(lambda i, j: ())),
                AffineMapAttr(
                    AffineMap(
                        2,
                        0,
                        (
                            AffineExpr.constant(0),
                            AffineExpr.dimension(0) * 3 + AffineExpr.dimension(1),
                        ),
                    )
                ),
            )
        ),
        ArrayAttr(
            (
                memref_stream.IteratorTypeAttr.parallel(),
                memref_stream.IteratorTypeAttr.parallel(),
            )
        ),
        ArrayAttr((index(2), index(3))),
        ArrayAttr(()),
    )

    with ImplicitBuilder(op.body) as (a, b, _c_init):
        c = arith.MuliOp(a, b).result
        memref_stream.YieldOp(c)

    op.verify()

    a = ShapedArray(TypedPtr.new_float64([1, 2, 3, 4, 5, 6]), [2, 3])
    b = 2
    c = ShapedArray(TypedPtr.new_float64([-1, -1, -1, -1, -1, -1]), [1, 6])

    interpreter.run_op(op, (a, b, c))

    assert c.data == [2, 4, 6, 8, 10, 12]


def test_memref_stream_generic_reduction():
    interpreter = Interpreter(ModuleOp([]))
    interpreter.register_implementations(MemRefStreamFunctions())
    interpreter.register_implementations(ArithFunctions())

    op = memref_stream.GenericOp(
        (
            TestSSAValue(MemRefType(i32, [3])),
            TestSSAValue(MemRefType(i32, [3])),
        ),
        (TestSSAValue(MemRefType(i32, [])),),
        (),
        Region(Block(arg_types=(i32, i32, i32))),
        ArrayAttr(
            (
                AffineMapAttr(AffineMap.identity(1)),
                AffineMapAttr(AffineMap.identity(1)),
                AffineMapAttr(AffineMap.from_callable(lambda d0: ())),
            )
        ),
        ArrayAttr((memref_stream.IteratorTypeAttr.reduction(),)),
        ArrayAttr((index(3),)),
        ArrayAttr(()),
    )

    with ImplicitBuilder(op.body) as (lhs, rhs, acc):
        sum = arith.MuliOp(lhs, rhs).result
        new_acc = arith.AddiOp(sum, acc).result
        memref_stream.YieldOp(new_acc)

    op.verify()

    a = ShapedArray(TypedPtr.new_float64([1, 2, 3]), [3])
    b = ShapedArray(TypedPtr.new_float64([4, 5, 6]), [3])
    c = ShapedArray(TypedPtr.new_float64([0]), [])

    interpreter.run_op(op, (a, b, c))

    assert c.data == [32]


def test_memref_stream_generic_imperfect_nesting():
    interpreter = Interpreter(ModuleOp([]))
    interpreter.register_implementations(MemRefStreamFunctions())
    interpreter.register_implementations(ArithFunctions())

    f32 = Float32Type()

    op = memref_stream.GenericOp(
        (
            TestSSAValue(MemRefType(f32, [3, 2])),
            TestSSAValue(MemRefType(f32, [2, 3])),
        ),
        (TestSSAValue(MemRefType(f32, [3, 3])),),
        (),
        Region(Block(arg_types=(f32, f32, f32))),
        ArrayAttr(
            (
                AffineMapAttr(AffineMap.from_callable(lambda n, m, k: (n, k))),
                AffineMapAttr(AffineMap.from_callable(lambda n, m, k: (k, m))),
                AffineMapAttr(AffineMap.from_callable(lambda n, m: (n, m))),
            )
        ),
        ArrayAttr(
            (
                memref_stream.IteratorTypeAttr.parallel(),
                memref_stream.IteratorTypeAttr.parallel(),
                memref_stream.IteratorTypeAttr.reduction(),
            )
        ),
        ArrayAttr((index(3), index(3), index(2))),
        ArrayAttr(()),
    )

    with ImplicitBuilder(op.body) as (lhs, rhs, acc):
        sum = arith.MulfOp(lhs, rhs).result
        new_acc = arith.AddfOp(sum, acc).result
        memref_stream.YieldOp(new_acc)

    op.verify()

    a = ShapedArray(TypedPtr.new_float32([1.0, 2.0, 3.0, 4.0, 5.0, 6.0]), [3, 2])
    b = ShapedArray(TypedPtr.new_float32([4.0, 3.0, 5.0, 1.0, 2.0, 8.0]), [2, 3])
    c = ShapedArray(TypedPtr.new_float32([0.0] * 9), [3, 3])

    interpreter.run_op(op, (a, b, c))
    assert c == ShapedArray(
        TypedPtr.new_float32([6.0, 7.0, 21.0, 16.0, 17.0, 47.0, 26.0, 27.0, 73.0]),
        [3, 3],
    )


def test_memref_stream_generic_reduction_with_initial_value():
    interpreter = Interpreter(ModuleOp([]))
    interpreter.register_implementations(MemRefStreamFunctions())
    interpreter.register_implementations(ArithFunctions())

    f32 = Float32Type()

    op = memref_stream.GenericOp(
        (
            TestSSAValue(MemRefType(f32, [3, 2])),
            TestSSAValue(MemRefType(f32, [2, 3])),
        ),
        (TestSSAValue(MemRefType(f32, [3, 3])),),
        (TestSSAValue(f32),),
        Region(Block(arg_types=(f32, f32, f32))),
        ArrayAttr(
            (
                AffineMapAttr(AffineMap.from_callable(lambda n, m, k: (n, k))),
                AffineMapAttr(AffineMap.from_callable(lambda n, m, k: (k, m))),
                AffineMapAttr(AffineMap.from_callable(lambda n, m: (n, m))),
            )
        ),
        ArrayAttr(
            (
                memref_stream.IteratorTypeAttr.parallel(),
                memref_stream.IteratorTypeAttr.parallel(),
                memref_stream.IteratorTypeAttr.reduction(),
            )
        ),
        ArrayAttr((index(3), index(3), index(2))),
        ArrayAttr((IntAttr(0),)),
    )

    with ImplicitBuilder(op.body) as (lhs, rhs, acc):
        sum = arith.MulfOp(lhs, rhs).result
        new_acc = arith.AddfOp(sum, acc).result
        memref_stream.YieldOp(new_acc)

    op.verify()

    a = ShapedArray(TypedPtr.new_float32([1.0, 2.0, 3.0, 4.0, 5.0, 6.0]), [3, 2])
    b = ShapedArray(TypedPtr.new_float32([4.0, 3.0, 5.0, 1.0, 2.0, 8.0]), [2, 3])
    c = ShapedArray(TypedPtr.new_float32([0.0] * 9), [3, 3])

    interpreter.run_op(op, (a, b, c, 0.5))
    assert c == ShapedArray(
        TypedPtr.new_float32([6.5, 7.5, 21.5, 16.5, 17.5, 47.5, 26.5, 27.5, 73.5]),
        [3, 3],
    )


def test_memref_stream_interleaved_reduction_with_initial_value():
    interpreter = Interpreter(ModuleOp([]))
    interpreter.register_implementations(MemRefStreamFunctions())
    interpreter.register_implementations(ArithFunctions())

    f32 = Float32Type()

    op = memref_stream.GenericOp(
        (
            TestSSAValue(MemRefType(f32, [3, 5])),
            TestSSAValue(MemRefType(f32, [5, 8])),
        ),
        (TestSSAValue(MemRefType(f32, [3, 8])),),
        (TestSSAValue(f32),),
        Region(
            Block(
                arg_types=(f32, f32, f32, f32, f32, f32, f32, f32, f32, f32, f32, f32)
            )
        ),
        ArrayAttr(
            (
                AffineMapAttr(AffineMap.from_callable(lambda n, m, k, j: (n, k))),
                AffineMapAttr(
                    AffineMap.from_callable(lambda n, m, k, j: (k, m * 4 + j))
                ),
                AffineMapAttr(AffineMap.from_callable(lambda n, m, j: (n, m * 4 + j))),
            )
        ),
        ArrayAttr(
            (
                memref_stream.IteratorTypeAttr.parallel(),
                memref_stream.IteratorTypeAttr.parallel(),
                memref_stream.IteratorTypeAttr.reduction(),
                memref_stream.IteratorTypeAttr.interleaved(),
            )
        ),
        ArrayAttr((index(3), index(2), index(5), index(4))),
        ArrayAttr((IntAttr(0),)),
    )

    with ImplicitBuilder(op.body) as (
        lhs0,
        lhs1,
        lhs2,
        lhs3,
        rhs0,
        rhs1,
        rhs2,
        rhs3,
        acc0,
        acc1,
        acc2,
        acc3,
    ):
        sum0 = arith.MulfOp(lhs0, rhs0).result
        sum1 = arith.MulfOp(lhs1, rhs1).result
        sum2 = arith.MulfOp(lhs2, rhs2).result
        sum3 = arith.MulfOp(lhs3, rhs3).result
        new_acc0 = arith.AddfOp(sum0, acc0).result
        new_acc1 = arith.AddfOp(sum1, acc1).result
        new_acc2 = arith.AddfOp(sum2, acc2).result
        new_acc3 = arith.AddfOp(sum3, acc3).result
        memref_stream.YieldOp(new_acc0, new_acc1, new_acc2, new_acc3)

    op.verify()

    a = ShapedArray(TypedPtr.new_float32([float(i) for i in range(3 * 5)]), [3, 5])
    b = ShapedArray(
        TypedPtr.new_float32([float(i) / 100 for i in range(5 * 8)]), [5, 8]
    )
    c = ShapedArray(TypedPtr.new_float32([-1000.0] * (3 * 5)), [3, 5])

    with pytest.raises(
        NotImplementedError,
        match="Interpreter for interleaved operations not yet implemented",
    ):
        interpreter.run_op(op, (a, b, c, 0.5))
    # assert c == ShapedArray(
    #     TypedPtr.new_float32([ 2.9 ,  3.  ,  3.1 ,  3.2 ,  3.3 ,  3.4 ,  3.5 ,  3.6 ,  6.9 ,
    #     7.25,  7.6 ,  7.95,  8.3 ,  8.65,  9.  ,  9.35, 10.9 , 11.5 ,
    #    12.1 , 12.7 , 13.3 , 13.9 , 14.5 , 15.1 ]),
    #     [3, 3],
    # )



from typing import ClassVar

import pytest

from xdsl.dialects.bufferization import (
    AllocTensorOp,
    CloneOp,
    TensorFromMemRefConstraint,
    ToTensorOp,
)
from xdsl.dialects.builtin import (
    AnyUnrankedMemRefTypeConstr,
    IndexType,
    IntegerType,
    MemRefType,
    TensorType,
    UnitAttr,
    UnrankedMemRefType,
    UnrankedTensorType,
    f64,
)
from xdsl.dialects.test import TestOp
from xdsl.ir import Attribute
from xdsl.irdl import (
    ConstraintContext,
    EqAttrConstraint,
    IRDLOperation,
    VarConstraint,
    irdl_op_definition,
    operand_def,
)
from xdsl.utils.exceptions import VerifyException


def test_tensor_from_memref_inference():
    constr = TensorFromMemRefConstraint(MemRefType.constr())
    assert not constr.can_infer(set())

    constr2 = TensorFromMemRefConstraint(
        EqAttrConstraint(MemRefType(f64, [10, 20, 30]))
    )
    assert constr2.can_infer(set())
    assert constr2.infer(ConstraintContext()) == TensorType(f64, [10, 20, 30])

    constr3 = TensorFromMemRefConstraint(
        EqAttrConstraint(UnrankedMemRefType.from_type(f64))
    )
    assert constr3.can_infer(set())
    assert constr3.infer(ConstraintContext()) == UnrankedTensorType(f64)


@irdl_op_definition
class TensorFromMemRefOp(IRDLOperation):
    name = "test.tensor_from_memref"
    T: ClassVar = VarConstraint("T", MemRefType.constr() | AnyUnrankedMemRefTypeConstr)

    in_tensor = operand_def(
        TensorFromMemRefConstraint(
            MemRefType.constr(element_type=EqAttrConstraint(IndexType()))
        )
    )

    in_var_memref = operand_def(T)

    in_var_tensor = operand_def(TensorFromMemRefConstraint(T))


def test_tensor_from_memref_constraint():
    [v_memref, v_tensor] = TestOp(
        result_types=[
            MemRefType(IndexType(), [10, 20, 30]),
            TensorType(IndexType(), [10, 20, 30]),
        ]
    ).res
    op1 = TensorFromMemRefOp(operands=(v_tensor, v_memref, v_tensor))
    op1.verify()

    [v_unranked_memref, v_unranked_tensor] = TestOp(
        result_types=[
            UnrankedMemRefType.from_type(IndexType()),
            UnrankedTensorType(IndexType()),
        ]
    ).res
    op2 = TensorFromMemRefOp(operands=(v_tensor, v_unranked_memref, v_unranked_tensor))
    op2.verify()


@pytest.mark.parametrize(
    "type1, type2, type3, error",
    [
        (
            IndexType(),
            MemRefType(IndexType(), [10, 20, 30]),
            TensorType(IndexType(), [10, 20, 30]),
            "Expected tensor or unranked tensor type, got index",
        ),
        (
            TensorType(IntegerType(32), [10, 10, 10]),
            MemRefType(IndexType(), [10, 20, 30]),
            TensorType(IndexType(), [10, 20, 30]),
            "Expected attribute index but got i32",
        ),
        (
            UnrankedTensorType(IndexType()),
            MemRefType(IndexType(), [10, 20, 30]),
            TensorType(IndexType(), [10, 20, 30]),
            "memref<\\*xindex> should be of base attribute memref",
        ),
        (
            TensorType(IndexType(), [10, 10, 10]),
            MemRefType(IndexType(), [10, 20, 30]),
            TensorType(IndexType(), [10, 20, 20]),
            "attribute memref<10x20x30xindex> expected from variable 'T', but got memref<10x20x20xindex>",
        ),
        (
            TensorType(IndexType(), [10, 10, 10]),
            MemRefType(IntegerType(32), [10, 20, 30]),
            TensorType(IndexType(), [10, 20, 30]),
            "attribute memref<10x20x30xi32> expected from variable 'T', but got memref<10x20x30xindex>",
        ),
    ],
)
def test_tensor_from_memref_constraint_failure(
    type1: Attribute, type2: Attribute, type3: Attribute, error: str
):
    [v1, v2, v3] = TestOp(
        result_types=[
            type1,
            type2,
            type3,
        ]
    ).res

    op1 = TensorFromMemRefOp(operands=(v1, v2, v3))
    with pytest.raises(VerifyException, match=error):
        op1.verify()


def test_to_tensor():
    memref_t = MemRefType(f64, [10, 20, 30])
    tensor_t = TensorType(f64, [10, 20, 30])
    memref_v = TestOp(result_types=[memref_t]).res[0]

    to_tensor = ToTensorOp(memref_v)
    assert to_tensor.memref == memref_v
    assert to_tensor.restrict is None
    assert to_tensor.writable is None
    assert to_tensor.tensor.type == tensor_t

    to_tensor = ToTensorOp(memref_v, writable=True, restrict=True)
    assert to_tensor.memref == memref_v
    assert to_tensor.restrict == UnitAttr()
    assert to_tensor.writable == UnitAttr()
    assert to_tensor.tensor.type == tensor_t


def test_alloc_tensor_static():
    t = TensorType(f64, [10, 20, 30])
    alloc_tensor = AllocTensorOp(t)

    assert alloc_tensor.tensor.type == t
    assert alloc_tensor.dynamic_sizes == ()
    assert alloc_tensor.copy is None
    assert alloc_tensor.size_hint is None


def test_clone():
    memref_t = MemRefType(f64, [10, 20, 30])
    memref_v = TestOp(result_types=[memref_t]).res[0]

    clone = CloneOp(memref_v)

    assert clone.input == memref_v
    assert clone.input.type == memref_t
    assert clone.output.type == memref_t



from math import copysign, isnan
from typing import cast

from xdsl.dialects import arith
from xdsl.dialects.builtin import FloatAttr, IntegerAttr
from xdsl.interpreter import (
    Interpreter,
    InterpreterFunctions,
    PythonValues,
    impl,
    register_impls,
)
from xdsl.irdl import base
from xdsl.utils.exceptions import InterpretationError
from xdsl.utils.isattr import isattr


@register_impls
class ArithFunctions(InterpreterFunctions):
    @impl(arith.ConstantOp)
    def run_constant(
        self, interpreter: Interpreter, op: arith.ConstantOp, args: PythonValues
    ) -> PythonValues:
        value = op.value
        interpreter.interpreter_assert(
            isattr(op.value, base(IntegerAttr) | base(FloatAttr)),
            f"arith.constant not implemented for {type(op.value)}",
        )
        value = cast(IntegerAttr, op.value)
        return (value.value.data,)

    @impl(arith.SubiOp)
    def run_subi(self, interpreter: Interpreter, op: arith.SubiOp, args: PythonValues):
        return (args[0] - args[1],)

    @impl(arith.AddiOp)
    def run_addi(self, interpreter: Interpreter, op: arith.AddiOp, args: PythonValues):
        return (args[0] + args[1],)

    @impl(arith.MuliOp)
    def run_muli(self, interpreter: Interpreter, op: arith.MuliOp, args: PythonValues):
        return (args[0] * args[1],)

    @impl(arith.SubfOp)
    def run_subf(self, interpreter: Interpreter, op: arith.SubfOp, args: PythonValues):
        return (args[0] - args[1],)

    @impl(arith.AddfOp)
    def run_addf(self, interpreter: Interpreter, op: arith.AddfOp, args: PythonValues):
        return (args[0] + args[1],)

    @impl(arith.MulfOp)
    def run_mulf(self, interpreter: Interpreter, op: arith.MulfOp, args: PythonValues):
        return (args[0] * args[1],)

    @impl(arith.MinimumfOp)
    def run_minimumf(
        self, interpreter: Interpreter, op: arith.MinimumfOp, args: PythonValues
    ):
        if isnan(args[0]) or isnan(args[1]):
            return (float("NaN"),)
        if args[0] == 0 and args[1] == 0:
            if copysign(1.0, args[0]) < 0 or copysign(1.0, args[1]) < 0:
                return (-0.0,)
            else:
                return (0.0,)
        return (min(args[0], args[1]),)

    @impl(arith.MaximumfOp)
    def run_maximumf(
        self, interpreter: Interpreter, op: arith.MaximumfOp, args: PythonValues
    ):
        if isnan(args[0]) or isnan(args[1]):
            return (float("NaN"),)
        if args[0] == 0 and args[1] == 0:
            if copysign(1.0, args[0]) > 0 or copysign(1.0, args[1]) > 0:
                return (0.0,)
            else:
                return (-0.0,)
        return (max(args[0], args[1]),)

    @impl(arith.CmpiOp)
    def run_cmpi(self, interpreter: Interpreter, op: arith.CmpiOp, args: PythonValues):
        match op.predicate.value.data:
            case 0:  # "eq"
                return (args[0] == args[1],)
            case 1:  # "ne"
                return (args[0] != args[1],)
            case 2:  # "slt"
                return (args[0] < args[1],)
            case 3:  # "sle"
                return (args[0] <= args[1],)
            case 4:  # "sgt"
                return (args[0] > args[1],)
            case 5:  # "sge"
                return (args[0] >= args[1],)
            case 6:  # "ult"
                return (args[0] < args[1],)
            case 7:  # "ule"
                return (args[0] <= args[1],)
            case 8:  # "ugt"
                return (args[0] > args[1],)
            case 9:  # "uge"
                return (args[0] >= args[1],)
            case _:
                raise InterpretationError(
                    f"arith.cmpi predicate {op.predicate} mot implemented yet."
                )



import pytest

from xdsl.context import Context
from xdsl.dialects import riscv
from xdsl.dialects.builtin import (
    IntAttr,
    IntegerAttr,
    ModuleOp,
    Signedness,
    i32,
)
from xdsl.parser import Parser
from xdsl.transforms.canonicalization_patterns.riscv import get_constant_value
from xdsl.utils.exceptions import ParseError, VerifyException
from xdsl.utils.test_value import TestSSAValue


def test_add_op():
    a1 = TestSSAValue(riscv.Registers.A1)
    a2 = TestSSAValue(riscv.Registers.A2)
    add_op = riscv.AddOp(a1, a2, rd=riscv.Registers.A0)
    a0 = add_op.rd

    assert a1.type is add_op.rs1.type
    assert a2.type is add_op.rs2.type
    assert isinstance(a0.type, riscv.IntRegisterType)
    assert isinstance(a1.type, riscv.IntRegisterType)
    assert isinstance(a2.type, riscv.IntRegisterType)
    assert a0.type.register_name.data == "a0"
    assert a0.type.index == IntAttr(10)
    assert a1.type.register_name.data == "a1"
    assert a1.type.index == IntAttr(11)
    assert a2.type.register_name.data == "a2"
    assert a2.type.index == IntAttr(12)

    # Registers that aren't predefined should not have an index.
    assert riscv.IntRegisterType.infinite_register(1).index == IntAttr(-2)


def test_is_non_zero():
    # Test zero register
    x0_reg = riscv.IntRegisterType.from_name("x0")
    assert not riscv.is_non_zero(riscv.Registers.ZERO)
    assert not riscv.is_non_zero(x0_reg)

    # Test non-zero registers
    a0_reg = riscv.Registers.A0
    t0_reg = riscv.Registers.T0
    assert riscv.is_non_zero(a0_reg)
    assert riscv.is_non_zero(t0_reg)

    # Test unallocated register
    unalloc_reg = riscv.IntRegisterType.unallocated()
    assert not riscv.is_non_zero(unalloc_reg)


def test_csr_op():
    a1 = TestSSAValue(riscv.Registers.A1)
    zero = TestSSAValue(riscv.Registers.ZERO)
    csr = IntegerAttr(16, i32)
    # CsrrwOp
    riscv.CsrrwOp(rs1=a1, csr=csr, rd=riscv.Registers.A2).verify()
    riscv.CsrrwOp(rs1=a1, csr=csr, rd=riscv.Registers.ZERO).verify()
    riscv.CsrrwOp(rs1=a1, csr=csr, writeonly=True, rd=riscv.Registers.ZERO).verify()
    with pytest.raises(VerifyException):
        riscv.CsrrwOp(rs1=a1, csr=csr, writeonly=True, rd=riscv.Registers.A2).verify()
    # CsrrsOp
    riscv.CsrrsOp(rs1=a1, csr=csr, rd=riscv.Registers.A2).verify()
    riscv.CsrrsOp(rs1=zero, csr=csr, rd=riscv.Registers.A2).verify()
    riscv.CsrrsOp(rs1=zero, csr=csr, readonly=True, rd=riscv.Registers.A2).verify()
    with pytest.raises(VerifyException):
        riscv.CsrrsOp(rs1=a1, csr=csr, readonly=True, rd=riscv.Registers.A2).verify()
    # CsrrcOp
    riscv.CsrrcOp(rs1=a1, csr=csr, rd=riscv.Registers.A2).verify()
    riscv.CsrrcOp(rs1=zero, csr=csr, rd=riscv.Registers.A2).verify()
    riscv.CsrrcOp(rs1=zero, csr=csr, readonly=True, rd=riscv.Registers.A2).verify()
    with pytest.raises(VerifyException):
        riscv.CsrrcOp(rs1=a1, csr=csr, readonly=True, rd=riscv.Registers.A2).verify()
    # CsrrwiOp
    riscv.CsrrwiOp(
        csr=csr, immediate=IntegerAttr(0, i32), rd=riscv.Registers.A2
    ).verify()
    riscv.CsrrwiOp(
        csr=csr, immediate=IntegerAttr(0, i32), rd=riscv.Registers.ZERO
    ).verify()
    riscv.CsrrwiOp(
        csr=csr, immediate=IntegerAttr(0, i32), writeonly=True, rd=riscv.Registers.ZERO
    ).verify()
    with pytest.raises(VerifyException):
        riscv.CsrrwiOp(
            csr=csr,
            immediate=IntegerAttr(0, i32),
            writeonly=True,
            rd=riscv.Registers.A2,
        ).verify()
    # CsrrsiOp
    riscv.CsrrsiOp(
        csr=csr, immediate=IntegerAttr(0, i32), rd=riscv.Registers.A2
    ).verify()
    riscv.CsrrsiOp(
        csr=csr, immediate=IntegerAttr(1, i32), rd=riscv.Registers.A2
    ).verify()
    # CsrrciOp
    riscv.CsrrciOp(
        csr=csr, immediate=IntegerAttr(0, i32), rd=riscv.Registers.A2
    ).verify()
    riscv.CsrrsiOp(
        csr=csr, immediate=IntegerAttr(1, i32), rd=riscv.Registers.A2
    ).verify()


def test_comment_op():
    comment_op = riscv.CommentOp("my comment")

    assert comment_op.comment.data == "my comment"

    code = riscv.riscv_code(ModuleOp([comment_op]))
    assert code == "    # my comment\n"


def test_label_op_without_comment():
    label_str = "mylabel"
    label_op = riscv.LabelOp(label_str)

    assert label_op.label.data == label_str

    code = riscv.riscv_code(ModuleOp([label_op]))
    assert code == f"{label_str}:\n"


def test_label_op_with_comment():
    label_str = "mylabel"
    label_op = riscv.LabelOp(f"{label_str}", comment="my label")

    assert label_op.label.data == label_str

    code = riscv.riscv_code(ModuleOp([label_op]))
    assert code == f"{label_str}:                                         # my label\n"


def test_return_op():
    return_op = riscv.EbreakOp(comment="my comment")

    assert return_op.comment is not None

    assert return_op.comment.data == "my comment"

    code = riscv.riscv_code(ModuleOp([return_op]))
    assert code == "    ebreak                                       # my comment\n"


def test_immediate_i_inst():
    # I-Type - 12-bits signed immediate
    lb, ub = Signedness.SIGNED.value_range(12)
    a1 = TestSSAValue(riscv.Registers.A1)

    with pytest.raises(VerifyException):
        riscv.AddiOp(a1, ub, rd=riscv.Registers.A0)

    with pytest.raises(VerifyException):
        riscv.AddiOp(a1, lb - 1, rd=riscv.Registers.A0)

    riscv.AddiOp(a1, ub - 1, rd=riscv.Registers.A0)
    riscv.AddiOp(a1, lb, rd=riscv.Registers.A0)


def test_immediate_s_inst():
    # S-Type - 12-bits signed immediate
    lb, ub = Signedness.SIGNED.value_range(12)
    a1 = TestSSAValue(riscv.Registers.A1)
    a2 = TestSSAValue(riscv.Registers.A2)

    with pytest.raises(VerifyException):
        riscv.SwOp(a1, a2, ub)

    with pytest.raises(VerifyException):
        riscv.SwOp(a1, a2, lb - 1)

    riscv.SwOp(a1, a2, ub - 1)
    riscv.SwOp(a1, a2, lb)


def test_immediate_u_j_inst():
    # U-Type and J-Type - 20-bits immediate
    lb, ub = Signedness.SIGNLESS.value_range(20)
    assert ub == 1048576
    assert lb == -524288

    with pytest.raises(VerifyException):
        riscv.LuiOp(ub)

    with pytest.raises(VerifyException):
        riscv.LuiOp(lb - 1)

    riscv.LuiOp(ub - 1)
    riscv.LuiOp(lb)


def test_immediate_jalr_inst():
    # Jalr - 12-bits immediate
    a1 = TestSSAValue(riscv.Registers.A1)

    with pytest.raises(VerifyException):
        riscv.JalrOp(a1, 1 << 12, rd=riscv.Registers.A0)

    with pytest.raises(VerifyException):
        riscv.JalrOp(a1, -(1 << 12) - 2, rd=riscv.Registers.A0)

    riscv.JalrOp(a1, (1 << 11) - 1, rd=riscv.Registers.A0)


def test_immediate_pseudo_inst():
    lb, ub = Signedness.SIGNLESS.value_range(32)
    assert ub == 4294967296
    assert lb == -2147483648

    # Pseudo-Instruction with custom handling
    with pytest.raises(VerifyException):
        riscv.LiOp(ub, rd=riscv.Registers.A0)

    with pytest.raises(VerifyException):
        riscv.LiOp(lb - 1, rd=riscv.Registers.A0)

    riscv.LiOp(ub - 1, rd=riscv.Registers.A0)
    riscv.LiOp(lb, rd=riscv.Registers.A0)


def test_immediate_shift_inst():
    # Shift instructions (SLLI, SRLI, SRAI) - 5-bits immediate
    a1 = TestSSAValue(riscv.Registers.A1)

    with pytest.raises(VerifyException):
        riscv.SlliOp(a1, 1 << 5, rd=riscv.Registers.A0)

    with pytest.raises(VerifyException):
        riscv.SlliOp(a1, -1, rd=riscv.Registers.A0)

    riscv.SlliOp(a1, (1 << 5) - 1, rd=riscv.Registers.A0)


def test_float_register():
    with pytest.raises(
        VerifyException, match="Invalid register name ft9 for register set RV32I."
    ):
        riscv.IntRegisterType.from_name("ft9")
    with pytest.raises(
        VerifyException, match="Invalid register name a0 for register set RV32F."
    ):
        riscv.FloatRegisterType.from_name("a0")

    a1 = TestSSAValue(riscv.Registers.A1)
    a2 = TestSSAValue(riscv.Registers.A2)
    with pytest.raises(VerifyException, match="Operation does not verify"):
        riscv.FAddSOp(a1, a2).verify()

    f1 = TestSSAValue(riscv.Registers.FT0)
    f2 = TestSSAValue(riscv.Registers.FT1)
    riscv.FAddSOp(f1, f2).verify()


def test_riscv_parse_immediate_value():
    ctx = Context()
    ctx.load_dialect(riscv.RISCV)

    prog = """riscv.jalr %0, 1.1, !riscv.reg : (!riscv.reg) -> ()"""
    parser = Parser(ctx, prog)
    with pytest.raises(ParseError, match="Expected immediate"):
        parser.parse_operation()


def test_asm_section():
    section = riscv.AssemblySectionOp("section")
    section.verify()


def test_get_constant_value():
    li_op = riscv.LiOp(1)
    li_val = get_constant_value(li_op.rd)
    assert li_val == IntegerAttr.from_int_and_width(1, 32)
    zero_op = riscv.GetRegisterOp(riscv.Registers.ZERO)
    zero_val = get_constant_value(zero_op.res)
    assert zero_val == IntegerAttr.from_int_and_width(0, 32)


def test_int_abi_name_by_index():
    assert riscv.IntRegisterType.abi_name_by_index() == {
        0: "zero",
        1: "ra",
        2: "sp",
        3: "gp",
        4: "tp",
        5: "t0",
        6: "t1",
        7: "t2",
        8: "s0",
        9: "s1",
        10: "a0",
        11: "a1",
        12: "a2",
        13: "a3",
        14: "a4",
        15: "a5",
        16: "a6",
        17: "a7",
        18: "s2",
        19: "s3",
        20: "s4",
        21: "s5",
        22: "s6",
        23: "s7",
        24: "s8",
        25: "s9",
        26: "s10",
        27: "s11",
        28: "t3",
        29: "t4",
        30: "t5",
        31: "t6",
    }


def test_int_from_index():
    assert riscv.IntRegisterType.from_index(0) == riscv.Registers.ZERO
    assert riscv.IntRegisterType.from_index(10) == riscv.Registers.A0
    assert riscv.IntRegisterType.from_index(20) == riscv.Registers.S4
    assert riscv.IntRegisterType.from_index(30) == riscv.Registers.T5

    with pytest.raises(KeyError):
        riscv.IntRegisterType.from_index(40)

    assert riscv.IntRegisterType.from_index(
        -10
    ) == riscv.IntRegisterType.infinite_register(9)


def test_float_abi_name_by_index():
    assert riscv.FloatRegisterType.abi_name_by_index() == {
        0: "ft0",
        1: "ft1",
        2: "ft2",
        3: "ft3",
        4: "ft4",
        5: "ft5",
        6: "ft6",
        7: "ft7",
        8: "fs0",
        9: "fs1",
        10: "fa0",
        11: "fa1",
        12: "fa2",
        13: "fa3",
        14: "fa4",
        15: "fa5",
        16: "fa6",
        17: "fa7",
        18: "fs2",
        19: "fs3",
        20: "fs4",
        21: "fs5",
        22: "fs6",
        23: "fs7",
        24: "fs8",
        25: "fs9",
        26: "fs10",
        27: "fs11",
        28: "ft8",
        29: "ft9",
        30: "ft10",
        31: "ft11",
    }


def test_float_from_index():
    assert riscv.FloatRegisterType.from_index(0) == riscv.Registers.FT0
    assert riscv.FloatRegisterType.from_index(10) == riscv.Registers.FA0
    assert riscv.FloatRegisterType.from_index(20) == riscv.Registers.FS4
    assert riscv.FloatRegisterType.from_index(30) == riscv.Registers.FT10

    with pytest.raises(KeyError):
        riscv.FloatRegisterType.from_index(40)

    assert riscv.FloatRegisterType.from_index(
        -10
    ) == riscv.FloatRegisterType.infinite_register(9)



from pathlib import Path
from typing import cast

from xdsl.parser import GenericParser, ParserState
from xdsl.utils.lexer import Input

from .lexer import ToyLexer, ToyToken, ToyTokenKind
from .location import loc
from .toy_ast import (
    BinaryExprAST,
    CallExprAST,
    ExprAST,
    FunctionAST,
    LiteralExprAST,
    ModuleAST,
    NumberExprAST,
    PrintExprAST,
    PrototypeAST,
    ReturnExprAST,
    VarDeclExprAST,
    VariableExprAST,
    VarType,
)


class ToyParser(GenericParser[ToyTokenKind]):
    def __init__(self, file: Path, program: str):
        super().__init__(ParserState(ToyLexer(Input(program, str(file)))))

    def getToken(self):
        """Returns current token in parser"""
        return self._current_token

    def getTokenPrecedence(self) -> int:
        """Returns precedence if the current token is a binary operation, -1 otherwise"""
        PRECEDENCE = {
            "-": 20,
            "+": 20,
            "*": 40,
        }
        op = self._current_token.text

        return PRECEDENCE.get(op, -1)

    def peek(self, pattern: str | ToyTokenKind) -> ToyToken | None:
        """
        Returns token matching pattern or None
        """
        token = self._current_token

        if isinstance(pattern, str):
            if token.text == pattern:
                return token
        else:
            if token.kind == pattern:
                return token

    def check(self, pattern: str | ToyTokenKind) -> bool:
        """
        Verifies that the current token fits the pattern,
        returns False otherwise
        """
        return self.peek(pattern) is not None

    def pop(self) -> ToyToken:
        return self._consume_token()

    def pop_pattern(self, pattern: str) -> ToyToken:
        """
        Verifies that the current token fits the pattern,
        raises ParseError otherwise
        """
        token = self._consume_token()
        if token.text != pattern:
            self.raise_error(f"Expected '{pattern}'", token.span.start, token.span.end)
        return token

    def pop_token(self, tokenType: ToyTokenKind) -> ToyToken:
        """
        Verifies that the current token is of expected type,
        raises ParseError otherwise
        """
        return self._consume_token(tokenType)

    def parseModule(self):
        """
        Parse a full Module. A module is a list of function definitions.
        """
        functions: list[FunctionAST] = []

        while not self.check(ToyTokenKind.EOF):
            functions.append(self.parseDefinition())

        # If we didn't reach EOF, there was an error during parsing
        self.pop_token(ToyTokenKind.EOF)

        return ModuleAST(tuple(functions))

    def parseReturn(self):
        """
        Parse a return statement.
        return :== return ; | return expr ;
        """
        returnToken = self.pop_pattern("return")
        expr = None

        # Return takes an optional argument
        if not self.check(";"):
            expr = self.parseExpression()

        return ReturnExprAST(loc(returnToken), expr)

    def parseNumberExpr(self):
        """
        Parse a literal number.
        numberexpr ::= number
        """
        numberToken = self.pop_token(ToyTokenKind.NUMBER)
        return NumberExprAST(loc(numberToken), float(numberToken.span.text))

    def parseTensorLiteralExpr(self) -> LiteralExprAST | NumberExprAST:
        """
        Parse a literal array expression.
        tensorLiteral ::= [ literalList ] | number
        literalList ::= tensorLiteral | tensorLiteral, literalList
        """
        if self.check(ToyTokenKind.NUMBER):
            return self.parseNumberExpr()

        openBracket = self.pop_pattern("[")

        # Hold the list of values at this nesting level.
        values: list[LiteralExprAST | NumberExprAST] = []
        # Hold the dimensions for all the nesting inside this level.
        dims: list[int] = []

        while True:
            # We can have either another nested array or a number literal.
            if self.check("["):
                values.append(self.parseTensorLiteralExpr())
            else:
                if not self.check(ToyTokenKind.NUMBER):
                    self.raise_error("Expected <num> or [ in literal expression")
                values.append(self.parseNumberExpr())

            # End of this list on ']'
            if self.check("]"):
                break

            # Elements are separated by a comma.
            self.pop_pattern(",")

        self.pop_pattern("]")

        # Fill in the dimensions now. First the current nesting level:
        dims.append(len(values))

        # If there is any nested array, process all of them and ensure
        # that dimensions are uniform.
        if any(type(val) is LiteralExprAST for val in values):
            allTensors = all(type(val) is LiteralExprAST for val in values)
            if not allTensors:
                self.raise_error(
                    "Expected uniform well-nested dimensions inside literal expression"
                )

            tensor_values = cast(list[LiteralExprAST], values)
            first = tensor_values[0].dims
            allEqual = all(val.dims == first for val in tensor_values)
            if not allEqual:
                self.raise_error(
                    "Expected uniform well-nested dimensions inside literal expression"
                )

            dims += first

        return LiteralExprAST(loc(openBracket), values, dims)

    def parseParenExpr(self) -> ExprAST:
        "parenexpr ::= '(' expression ')'"
        self.pop_pattern("(")
        v = self.parseExpression()
        self.pop_pattern(")")
        return v

    def parseIdentifierExpr(self):
        """
        identifierexpr
        ::= identifier
        ::= identifier '(' expression ')'
        """
        name = self.pop_token(ToyTokenKind.IDENTIFIER)
        if not self.check("("):
            # Simple variable ref.
            return VariableExprAST(loc(name), name.text)

        # This is a function call.
        self.pop_pattern("(")
        args: list[ExprAST] = []
        while True:
            args.append(self.parseExpression())
            if self.check(")"):
                break
            self.pop_pattern(",")
        self.pop_pattern(")")

        if name.text == "print":
            # It can be a builtin call to print
            if len(args) != 1:
                self.raise_error("Expected <single arg> as argument to print()")

            return PrintExprAST(loc(name), args[0])

        return CallExprAST(loc(name), name.text, args)

    def parsePrimary(self) -> ExprAST | None:
        """
        primary
        ::= identifierexpr
        ::= numberexpr
        ::= parenexpr
        ::= tensorliteral
        """
        current = self._current_token
        if current.kind == ToyTokenKind.IDENTIFIER:
            return self.parseIdentifierExpr()
        elif current.kind == ToyTokenKind.NUMBER:
            return self.parseNumberExpr()
        elif current.text == "(":
            return self.parseParenExpr()
        elif current.text == "[":
            return self.parseTensorLiteralExpr()
        elif current.text == ";":
            return None
        elif current.text == "}":
            return None
        else:
            self.raise_error("Expected expression or one of `;`, `}`")

    def parsePrimaryNotNone(self) -> ExprAST:
        """
        primary
        ::= identifierexpr
        ::= numberexpr
        ::= parenexpr
        ::= tensorliteral
        """
        current = self._current_token
        if current.kind == ToyTokenKind.IDENTIFIER:
            return self.parseIdentifierExpr()
        elif current.kind == ToyTokenKind.NUMBER:
            return self.parseNumberExpr()
        elif current.text == "(":
            return self.parseParenExpr()
        elif current.text == "[":
            return self.parseTensorLiteralExpr()
        else:
            self.raise_error("Expected expression")

    def parseBinOpRHS(self, exprPrec: int, lhs: ExprAST) -> ExprAST:
        """
        Recursively parse the right hand side of a binary expression, the ExprPrec
        argument indicates the precedence of the current binary operator.

        binoprhs ::= ('+' primary)*
        """
        # If this is a binop, find its precedence.
        while True:
            tokPrec = self.getTokenPrecedence()

            # If this is a binop that binds at least as tightly as the current binop,
            # consume it, otherwise we are done.
            if tokPrec < exprPrec:
                return lhs

            # Okay, we know this is a binop.
            binOp = self.pop_token(ToyTokenKind.OPERATOR).text

            # Parse the primary expression after the binary operator.
            rhs = self.parsePrimary()

            if rhs is None:
                self.raise_error("Expected expression to complete binary operator")

            # If BinOp binds less tightly with rhs than the operator after rhs, let
            # the pending operator take rhs as its lhs.
            nextPrec = self.getTokenPrecedence()
            if tokPrec < nextPrec:
                rhs = self.parseBinOpRHS(tokPrec + 1, rhs)

            # Merge lhs/rhs
            lhs = BinaryExprAST(rhs.loc, binOp, lhs, rhs)

    def parseExpression(self) -> ExprAST:
        """expression::= primary binop rhs"""
        lhs = self.parsePrimaryNotNone()
        return self.parseBinOpRHS(0, lhs)

    def parseType(self):
        """
        type ::= < shape_list >
        shape_list ::= num | num , shape_list
        """
        self.pop_pattern("<")
        shape: list[int] = []

        while token := self.pop_token(ToyTokenKind.NUMBER):
            shape.append(int(token.span.text))
            if self.parse_optional_characters(">"):
                break
            self.pop_pattern(",")

        return VarType(shape)

    def parseDeclaration(self):
        """
        Parse a variable declaration, it starts with a `var` keyword followed by
        and identifier and an optional type (shape specification) before the
        initializer.
        decl ::= var identifier [ type ] = expr
        """
        var = self.pop_pattern("var")
        name = self.pop_token(ToyTokenKind.IDENTIFIER).text

        # Type is optional, it can be inferred
        if self.check("<"):
            varType = self.parseType()
        else:
            varType = VarType([])

        self.pop_pattern("=")

        expr = self.parseExpression()
        return VarDeclExprAST(loc(var), name, varType, expr)

    def parseBlock(self) -> tuple[ExprAST, ...]:
        """
        Parse a block: a list of expression separated by semicolons and wrapped in
        curly braces.

        block ::= { expression_list }
        expression_list ::= block_expr ; expression_list
        block_expr ::= decl | "return" | expr
        """
        self.pop_pattern("{")
        exprList: list[ExprAST] = []

        # Ignore empty expressions: swallow sequences of semicolons.
        while self.check(";"):
            self.pop_pattern(";")

        while not self.check("}"):
            if self.check("var"):
                # Variable declaration
                exprList.append(self.parseDeclaration())
            elif self.check("return"):
                # Return statement
                exprList.append(self.parseReturn())
            else:
                # General expression
                exprList.append(self.parseExpression())

            # Ensure that elements are separated by a semicolon.
            self.pop_pattern(";")

            # Ignore empty expressions: swallow sequences of semicolons.
            while self.check(";"):
                self.pop_pattern(";")

        self.pop_pattern("}")

        return tuple(exprList)

    def parsePrototype(self):
        """
        prototype ::= def id '(' decl_list ')'
        decl_list ::= identifier | identifier, decl_list
        """
        defToken = self.pop_pattern("def")
        fnName = self.pop_token(ToyTokenKind.IDENTIFIER).text
        self.pop_pattern("(")

        args: list[VariableExprAST] = []
        if not self.check(")"):
            while True:
                arg = self.pop_token(ToyTokenKind.IDENTIFIER)
                args.append(VariableExprAST(loc(arg), arg.text))
                if not self.check(","):
                    break
                self.pop_pattern(",")

        self.pop_pattern(")")
        return PrototypeAST(loc(defToken), fnName, args)

    def parseDefinition(self):
        """
        Parse a function definition, we expect a prototype initiated with the
        `def` keyword, followed by a block containing a list of expressions.

        definition ::= prototype block
        """
        proto = self.parsePrototype()
        block = self.parseBlock()
        return FunctionAST(proto.loc, proto, block)



from __future__ import annotations

import abc
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterator, Sequence, Set
from dataclasses import KW_ONLY, dataclass, field
from inspect import isclass
from typing import TYPE_CHECKING, Generic, TypeAlias, TypeVar, cast

from typing_extensions import assert_never

from xdsl.ir import (
    Attribute,
    AttributeCovT,
    AttributeInvT,
    ParametrizedAttribute,
    TypedAttribute,
)
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.runtime_final import is_runtime_final

if TYPE_CHECKING:
    from xdsl.irdl import IRDLAttrConstraint


@dataclass
class ConstraintContext:
    """
    Contains the assignment of constraint variables.
    """

    _variables: dict[str, Attribute] = field(default_factory=dict)
    """The assignment of constraint variables."""

    _range_variables: dict[str, tuple[Attribute, ...]] = field(default_factory=dict)
    """The assignment of constraint range variables."""

    _int_variables: dict[str, int] = field(default_factory=dict)
    """The assignment of constraint int variables."""

    def get_variable(self, key: str) -> Attribute | None:
        return self._variables.get(key)

    def get_range_variable(self, key: str) -> tuple[Attribute, ...] | None:
        return self._range_variables.get(key)

    def get_int_variable(self, key: str) -> int | None:
        return self._int_variables.get(key)

    def set_variable(self, key: str, attr: Attribute):
        self._variables[key] = attr

    def set_range_variable(self, key: str, attr: tuple[Attribute, ...]):
        self._range_variables[key] = attr

    def set_int_variable(self, key: str, i: int):
        self._int_variables[key] = i

    @property
    def variables(self) -> Set[str]:
        return self._variables.keys()

    @property
    def range_variables(self) -> Set[str]:
        return self._range_variables.keys()

    @property
    def int_variables(self) -> Set[str]:
        return self._int_variables.keys()

    def copy(self):
        return ConstraintContext(
            self._variables.copy(),
            self._range_variables.copy(),
            self._int_variables.copy(),
        )

    def update(self, other: ConstraintContext):
        self._variables.update(other._variables)
        self._range_variables.update(other._range_variables)
        self._int_variables.update(other._int_variables)


_AttributeCovT = TypeVar("_AttributeCovT", bound=Attribute, covariant=True)

ConstraintVariableType: TypeAlias = Attribute | Sequence[Attribute] | int
"""
Possible types that a constraint variable can have.
"""


_T = TypeVar("_T")


@dataclass(frozen=True)
class VarExtractor(Generic[_T], abc.ABC):
    """
    A VarExtractor describes how to extract the value of a constraint variable
    from a value of type _T. The result of the extraction must be a subtype of
    ConstraintVariableType.
    """

    @abstractmethod
    def extract_var(self, a: _T) -> ConstraintVariableType: ...

    def flatten(self) -> Iterator[VarExtractor[_T]]:
        """Helper function to merge extractors, should only be overloaded by MergeExtractor"""
        yield self


@dataclass(frozen=True)
class MergeExtractor(VarExtractor[_T]):
    variable: str
    extractors: tuple[VarExtractor[_T], ...]
    """Extractors to merge, should not be empty."""

    def extract_var(self, a: _T) -> ConstraintVariableType:
        res = set(x.extract_var(a) for x in self.extractors)
        if len(res) == 1:
            return res.pop()
        else:
            raise ValueError(
                f"Value of variable {self.variable} could not be uniquely extracted.\n"
                f"Possible values are: {{{', '.join(sorted(str(x) for x in res))}}}"
            )

    def flatten(self) -> Iterator[VarExtractor[_T]]:
        return iter(self.extractors)


def merge_extractors(variable: str, *extractors: VarExtractor[_T]) -> VarExtractor[_T]:
    """
    Helper method used for merging dictionaries of extractors.
    Must not be called with no extractors.
    """
    if len(extractors) == 1:
        return extractors[0]
    extractor_tuple = tuple(x for ex in extractors for x in ex.flatten())
    return MergeExtractor(variable, extractor_tuple)


def merge_extractor_dicts(
    *dicts: dict[str, VarExtractor[_T]],
) -> dict[str, VarExtractor[_T]]:
    """
    Merge dictionaries of extractors.
    The keys of the resulting dictionary is equal to the union of the keys
    of the input dictionaries, with duplicated values being merged to a
    single MergeExtractor
    """
    keys = set(key for d in dicts for key in d)
    return {v: merge_extractors(v, *(d[v] for d in dicts if v in d)) for v in keys}


@dataclass(frozen=True)
class GenericAttrConstraint(Generic[AttributeCovT], ABC):
    """Constrain an attribute to a certain value."""

    @abstractmethod
    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        """
        Check if the attribute satisfies the constraint,
        or raise an exception otherwise.
        """
        ...

    def get_variable_extractors(self) -> dict[str, VarExtractor[AttributeCovT]]:
        """
        Get a dictionary of constraint variables to extractors for these variables,
        which provide a method to obtain the value of each constraint variable from
        the value of the attribute which is verified by this constraint.
        """
        return {}

    def can_infer(self, var_constraint_names: Set[str]) -> bool:
        """
        Check if there is enough information to infer the attribute given the
        constraint variables that are already set.
        """
        # By default, we cannot infer anything.
        return False

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        """
        Infer the attribute given the the values for all variables.

        Raises an exception if the attribute cannot be inferred. If `can_infer`
        returns `True` with the given constraint variables, this method should
        not raise an exception.
        """
        raise ValueError("Cannot infer attribute from constraint")

    def get_unique_base(self) -> type[Attribute] | None:
        """Get the unique base type that can satisfy the constraint, if any."""
        return None

    def __or__(
        self, value: GenericAttrConstraint[_AttributeCovT], /
    ) -> AnyOf[AttributeCovT | _AttributeCovT]:
        return AnyOf((self, value))

    def __and__(
        self, value: GenericAttrConstraint[AttributeCovT], /
    ) -> AllOf[AttributeCovT]:
        return AllOf((self, value))


AttrConstraint: TypeAlias = GenericAttrConstraint[Attribute]
ConstraintVariableTypeT = TypeVar(
    "ConstraintVariableTypeT", bound=ConstraintVariableType
)


@dataclass(frozen=True)
class IdExtractor(VarExtractor[ConstraintVariableTypeT]):
    """
    An extractor that simply returns its input.
    """

    def extract_var(self, a: ConstraintVariableTypeT) -> ConstraintVariableType:
        return a


TypedAttributeCovT = TypeVar("TypedAttributeCovT", bound=TypedAttribute, covariant=True)
TypedAttributeT = TypeVar("TypedAttributeT", bound=TypedAttribute)


@dataclass(frozen=True)
class TypedAttributeConstraint(GenericAttrConstraint[TypedAttributeCovT]):
    """
    Constrains the type of a typed attribute.
    """

    attr_constraint: GenericAttrConstraint[TypedAttributeCovT]

    type_constraint: GenericAttrConstraint[Attribute]

    def verify(self, attr: Attribute, constraint_context: ConstraintContext) -> None:
        self.attr_constraint.verify(attr, constraint_context)
        if not isinstance(attr, TypedAttribute):
            raise VerifyException(f"attribute {attr} expected to be a TypedAttribute")
        self.type_constraint.verify(attr.get_type(), constraint_context)

    @dataclass(frozen=True)
    class _Extractor(VarExtractor[TypedAttributeT]):
        inner: VarExtractor[Attribute]

        def extract_var(self, a: TypedAttributeT) -> ConstraintVariableType:
            return self.inner.extract_var(a.get_type())

    def get_variable_extractors(self) -> dict[str, VarExtractor[TypedAttributeCovT]]:
        return merge_extractor_dicts(
            self.attr_constraint.get_variable_extractors(),
            {
                v: self._Extractor(r)
                for v, r in self.type_constraint.get_variable_extractors().items()
            },
        )

    def can_infer(self, var_constraint_names: Set[str]) -> bool:
        return self.attr_constraint.can_infer(var_constraint_names)

    def infer(self, context: ConstraintContext) -> TypedAttributeCovT:
        return self.attr_constraint.infer(context)


@dataclass(frozen=True)
class VarConstraint(GenericAttrConstraint[AttributeCovT]):
    """
    Constrain an attribute with the given constraint, and constrain all occurences
    of this constraint (i.e, sharing the same name) to be equal.
    """

    name: str
    """The variable name. All uses of that name refer to the same variable."""

    constraint: GenericAttrConstraint[AttributeCovT]
    """The constraint that the variable must satisfy."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        ctx_attr = constraint_context.get_variable(self.name)
        if ctx_attr is not None:
            if attr != ctx_attr:
                raise VerifyException(
                    f"attribute {constraint_context.get_variable(self.name)} expected from variable "
                    f"'{self.name}', but got {attr}"
                )
        else:
            self.constraint.verify(attr, constraint_context)
            constraint_context.set_variable(self.name, attr)

    def get_variable_extractors(self) -> dict[str, VarExtractor[AttributeCovT]]:
        return merge_extractor_dicts(
            {self.name: IdExtractor()}, self.constraint.get_variable_extractors()
        )

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        v = context.get_variable(self.name)
        return cast(AttributeCovT, v)

    def can_infer(self, var_constraint_names: Set[str]) -> bool:
        return self.name in var_constraint_names

    def get_unique_base(self) -> type[Attribute] | None:
        return self.constraint.get_unique_base()


@dataclass(frozen=True, init=True)
class ConstraintVar:
    """
    Annotation used in PyRDL to define a constraint variable.
    For instance, the following code defines a constraint variable T,
    that can then be used in PyRDL:
    ```python
    T = Annotated[PyRDLConstraint, ConstraintVar("T")]
    ```
    """

    name: str
    """The variable name. All uses of that name refer to the same variable."""


@dataclass(frozen=True)
class EqAttrConstraint(Generic[AttributeCovT], GenericAttrConstraint[AttributeCovT]):
    """Constrain an attribute to be equal to another attribute."""

    attr: AttributeCovT
    """The attribute we want to check equality with."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        if attr != self.attr:
            raise VerifyException(f"Expected attribute {self.attr} but got {attr}")

    def can_infer(self, var_constraint_names: Set[str]) -> bool:
        return True

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        return self.attr

    def get_unique_base(self) -> type[Attribute] | None:
        return type(self.attr)


@dataclass(frozen=True)
class BaseAttr(Generic[AttributeCovT], GenericAttrConstraint[AttributeCovT]):
    """Constrain an attribute to be of a given base type."""

    attr: type[AttributeCovT]
    """The expected attribute base type."""

    def __repr__(self):
        return f"BaseAttr({self.attr.__name__})"

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        if not isinstance(attr, self.attr):
            raise VerifyException(
                f"{attr} should be of base attribute {self.attr.name}"
            )

    def can_infer(self, var_constraint_names: Set[str]) -> bool:
        return (
            is_runtime_final(self.attr)
            and issubclass(self.attr, ParametrizedAttribute)
            and not self.attr.get_irdl_definition().parameters
        )

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        assert issubclass(self.attr, ParametrizedAttribute)
        attr = self.attr.new(())
        return attr

    def get_unique_base(self) -> type[Attribute] | None:
        if is_runtime_final(self.attr):
            return self.attr
        return None


def attr_constr_coercion(
    attr: AttributeCovT | type[AttributeCovT] | GenericAttrConstraint[AttributeCovT],
) -> GenericAttrConstraint[AttributeCovT]:
    """
    Attributes are coerced into EqAttrConstraints,
    and Attribute types are coerced into BaseAttr.
    """
    if isinstance(attr, GenericAttrConstraint):
        return attr
    if isinstance(attr, Attribute):
        return EqAttrConstraint(attr)
    if isclass(attr):
        return BaseAttr(attr)
    assert_never(attr)


@dataclass(frozen=True)
class AnyAttr(GenericAttrConstraint[Attribute]):
    """Constraint that is verified by all attributes."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        pass


@dataclass(frozen=True, init=False)
class AnyOf(Generic[AttributeCovT], GenericAttrConstraint[AttributeCovT]):
    """Ensure that an attribute satisfies one of the given constraints."""

    attr_constrs: tuple[GenericAttrConstraint[AttributeCovT], ...]
    """The list of constraints that are checked."""

    def __init__(
        self,
        attr_constrs: Sequence[
            AttributeCovT | type[AttributeCovT] | GenericAttrConstraint[AttributeCovT]
        ],
    ):
        constrs: tuple[GenericAttrConstraint[AttributeCovT], ...] = tuple(
            attr_constr_coercion(constr) for constr in attr_constrs
        )
        object.__setattr__(
            self,
            "attr_constrs",
            constrs,
        )

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext | None = None,
    ) -> None:
        constraint_context = constraint_context or ConstraintContext()
        for attr_constr in self.attr_constrs:
            # Copy the constraint to ensure that if the constraint fails, the
            # constraint context is not modified.
            constraint_context_copy = constraint_context.copy()
            try:
                attr_constr.verify(attr, constraint_context_copy)
                # If the constraint succeeds, we update back the constraint variables
                constraint_context.update(constraint_context_copy)
                return
            except VerifyException:
                pass
        raise VerifyException(f"Unexpected attribute {attr}")

    def __or__(
        self, value: GenericAttrConstraint[_AttributeCovT], /
    ) -> AnyOf[AttributeCovT | _AttributeCovT]:
        return AnyOf((*self.attr_constrs, value))

    def get_variable_extractors(self) -> dict[str, VarExtractor[AttributeCovT]]:
        if len(self.attr_constrs) == 1:
            return self.attr_constrs[0].get_variable_extractors()
        return dict()

    def get_unique_base(self) -> type[Attribute] | None:
        bases = [constr.get_unique_base() for constr in self.attr_constrs]
        if None in bases:
            return None
        if len(set(bases)) == 1:
            return bases[0]
        return None


@dataclass(frozen=True)
class AllOf(GenericAttrConstraint[AttributeCovT]):
    """Ensure that an attribute satisfies all the given constraints."""

    attr_constrs: tuple[GenericAttrConstraint[AttributeCovT], ...]
    """The list of constraints that are checked."""

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        exc_bucket: list[VerifyException] = []

        for attr_constr in self.attr_constrs:
            try:
                attr_constr.verify(attr, constraint_context)
            except VerifyException as e:
                exc_bucket.append(e)

        if len(exc_bucket):
            if len(exc_bucket) == 1:
                raise VerifyException(str(exc_bucket[0])) from exc_bucket[0]
            exc_msg = "The following constraints were not satisfied:\n"
            exc_msg += "\n".join([str(e) for e in exc_bucket])
            raise VerifyException(exc_msg)

    def get_variable_extractors(self) -> dict[str, VarExtractor[AttributeCovT]]:
        return merge_extractor_dicts(
            *(constr.get_variable_extractors() for constr in self.attr_constrs)
        )

    def can_infer(self, var_constraint_names: Set[str]) -> bool:
        return any(
            constr.can_infer(var_constraint_names) for constr in self.attr_constrs
        )

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        for constr in self.attr_constrs:
            if constr.can_infer(context.variables):
                return constr.infer(context)
        raise ValueError("Cannot infer attribute from constraint")

    def get_unique_base(self) -> type[Attribute] | None:
        # This could be improved if we keep track of all the possible base types for
        # each constraint.
        for constr in self.attr_constrs:
            base = constr.get_unique_base()
            if base is not None:
                return base
        return None

    def __and__(
        self, value: GenericAttrConstraint[AttributeCovT], /
    ) -> AllOf[AttributeCovT]:
        return AllOf((*self.attr_constrs, value))


ParametrizedAttributeT = TypeVar("ParametrizedAttributeT", bound=ParametrizedAttribute)
ParametrizedAttributeCovT = TypeVar(
    "ParametrizedAttributeCovT", bound=ParametrizedAttribute, covariant=True
)


@dataclass(frozen=True, init=False)
class ParamAttrConstraint(
    Generic[ParametrizedAttributeCovT], GenericAttrConstraint[ParametrizedAttributeCovT]
):
    """
    Constrain an attribute to be of a given type,
    and also constrain its parameters with additional constraints.
    """

    base_attr: type[ParametrizedAttributeCovT]
    """The base attribute type."""

    param_constrs: tuple[AttrConstraint, ...]
    """The attribute parameter constraints"""

    def __init__(
        self,
        base_attr: type[ParametrizedAttributeCovT],
        param_constrs: Sequence[IRDLAttrConstraint | None],
    ):
        from xdsl.irdl import irdl_to_attr_constraint

        constrs = tuple(
            irdl_to_attr_constraint(constr) if constr is not None else AnyAttr()
            for constr in param_constrs
        )
        object.__setattr__(self, "base_attr", base_attr)
        object.__setattr__(self, "param_constrs", constrs)

    def __repr__(self):
        return f"ParamAttrConstraint({self.base_attr.__name__}, {repr(self.param_constrs)})"

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        if not isinstance(attr, self.base_attr):
            raise VerifyException(
                f"{attr} should be of base attribute {self.base_attr.name}"
            )
        if len(self.param_constrs) != len(attr.parameters):
            raise VerifyException(
                f"{len(self.param_constrs)} parameters expected, "
                f"but got {len(attr.parameters)}"
            )
        for idx, param_constr in enumerate(self.param_constrs):
            param_constr.verify(attr.parameters[idx], constraint_context)

    @dataclass(frozen=True)
    class _Extractor(VarExtractor[ParametrizedAttributeT]):
        idx: int
        inner: VarExtractor[Attribute]

        def extract_var(self, a: ParametrizedAttributeT) -> ConstraintVariableType:
            return self.inner.extract_var(a.parameters[self.idx])

    def get_variable_extractors(
        self,
    ) -> dict[str, VarExtractor[ParametrizedAttributeCovT]]:
        dicts: Generator[dict[str, VarExtractor[ParametrizedAttributeCovT]]] = (
            {
                v: self._Extractor(i, r)
                for v, r in param_constr.get_variable_extractors().items()
            }
            for i, param_constr in enumerate(self.param_constrs)
        )
        return merge_extractor_dicts(*dicts)

    def can_infer(self, var_constraint_names: Set[str]) -> bool:
        return is_runtime_final(self.base_attr) and all(
            constr.can_infer(var_constraint_names) for constr in self.param_constrs
        )

    def infer(self, context: ConstraintContext) -> ParametrizedAttributeCovT:
        params = tuple(constr.infer(context) for constr in self.param_constrs)
        attr = self.base_attr.new(params)
        return attr

    def get_unique_base(self) -> type[Attribute] | None:
        if is_runtime_final(self.base_attr):
            return self.base_attr
        return None


@dataclass(frozen=True, init=False)
class MessageConstraint(GenericAttrConstraint[AttributeCovT]):
    """
    Attach a message to a constraint, to provide more context when the constraint
    is not satisfied.
    """

    constr: GenericAttrConstraint[AttributeCovT]
    message: str

    def __init__(
        self,
        constr: (
            GenericAttrConstraint[AttributeCovT] | AttributeCovT | type[AttributeCovT]
        ),
        message: str,
    ):
        object.__setattr__(self, "constr", attr_constr_coercion(constr))
        object.__setattr__(self, "message", message)

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        try:
            return self.constr.verify(attr, constraint_context)
        except VerifyException as e:
            raise VerifyException(
                f"{self.message}\nUnderlying verification failure: {e.args[0]}",
                *e.args[1:],
            )

    def get_variable_extractors(self) -> dict[str, VarExtractor[AttributeCovT]]:
        return self.constr.get_variable_extractors()

    def get_unique_base(self) -> type[Attribute] | None:
        return self.constr.get_unique_base()

    def can_infer(self, var_constraint_names: Set[str]) -> bool:
        return self.constr.can_infer(var_constraint_names)

    def infer(self, context: ConstraintContext) -> AttributeCovT:
        return self.constr.infer(context)


@dataclass(frozen=True)
class IntConstraint(ABC):
    """Constrain an integer to certain values."""

    @abstractmethod
    def verify(
        self,
        i: int,
        constraint_context: ConstraintContext,
    ) -> None:
        """
        Check if the integer satisfies the constraint, or raise an exception otherwise.
        """
        ...

    def get_length_extractors(
        self,
    ) -> dict[str, VarExtractor[int]]:
        """
        Get a dictionary of variables that can be solved from this attribute.
        """
        return dict()

    def can_infer(self, var_constraint_names: Set[str]) -> bool:
        """
        Check if there is enough information to infer the integer given the
        constraint variables that are already set.
        """
        # By default, we cannot infer anything.
        return False

    def infer(self, context: ConstraintContext) -> int:
        """
        Infer the attribute given the the values for all variables.

        Raises an exception if the attribute cannot be inferred. If `can_infer`
        returns `True` with the given constraint variables, this method should
        not raise an exception.
        """
        raise ValueError("Cannot infer attribute from constraint")


class AnyInt(IntConstraint):
    """
    Constraint that is verified by all integers.
    """

    def verify(self, i: int, constraint_context: ConstraintContext) -> None:
        pass


@dataclass(frozen=True)
class IntVarConstraint(IntConstraint):
    """
    Constrain an integer with the given constraint, and constrain all occurences
    of this constraint (i.e, sharing the same name) to be equal.
    """

    name: str
    """The variable name. All uses of that name refer to the same variable."""

    constraint: IntConstraint
    """The constraint that the variable must satisfy."""

    def verify(
        self,
        i: int,
        constraint_context: ConstraintContext,
    ) -> None:
        if self.name in constraint_context.int_variables:
            if i != constraint_context.get_int_variable(self.name):
                raise VerifyException(
                    f"integer {constraint_context.get_int_variable(self.name)} expected from int variable "
                    f"'{self.name}', but got {i}"
                )
        else:
            self.constraint.verify(i, constraint_context)
            constraint_context.set_int_variable(self.name, i)

    def get_length_extractors(
        self,
    ) -> dict[str, VarExtractor[int]]:
        return {self.name: IdExtractor()}

    def can_infer(self, var_constraint_names: Set[str]) -> bool:
        return self.name in var_constraint_names

    def infer(
        self,
        context: ConstraintContext,
    ) -> int:
        v = context.get_int_variable(self.name)
        assert isinstance(v, int)
        return v


@dataclass(frozen=True)
class GenericRangeConstraint(Generic[AttributeCovT], ABC):
    """Constrain a range of attributes to certain values."""

    @abstractmethod
    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        """
        Check if the range satisfies the constraint, or raise an exception otherwise.
        The range can contain Nones, which represent an attribute not to be checked.
        """
        ...

    def get_variable_extractors(
        self,
    ) -> dict[str, VarExtractor[Sequence[AttributeCovT]]]:
        """
        Get a dictionary of constraint variables to extractors for these variables,
        which provide a method to obtain the value of each constraint variable from
        the value of the attribute range which is verified by this constraint.
        """
        return {}

    def get_length_extractors(
        self,
    ) -> dict[str, VarExtractor[int]]:
        """
        Get a dictionary of variables that can be solved using the length of the range.
        """
        return dict()

    def can_infer(self, var_constraint_names: Set[str], *, length_known: bool) -> bool:
        """
        Check if there is enough information to infer the attribute given the
        constraint variables that are already set, and whether the length of the
        range is known in advance.
        """
        # By default, we cannot infer anything.
        return False

    def infer(
        self, context: ConstraintContext, *, length: int | None
    ) -> Sequence[AttributeCovT]:
        """
        Infer the attribute given the the values for all variables, and possibly
        the length of the range if known.

        Raises an exception if the attribute cannot be inferred. If `can_infer`
        returns `True` with the given constraint variables, this method should
        not raise an exception.
        """
        raise ValueError("Cannot infer attribute from constraint")


RangeConstraint: TypeAlias = GenericRangeConstraint[Attribute]


@dataclass(frozen=True)
class RangeVarConstraint(GenericRangeConstraint[AttributeCovT]):
    """
    Constrain an attribute range with the given constraint, and constrain all occurences
    of this constraint (i.e, sharing the same name) to be equal.
    """

    name: str
    """The variable name. All uses of that name refer to the same variable."""

    constraint: GenericRangeConstraint[AttributeCovT]
    """The constraint that the variable must satisfy."""

    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        ctx_attrs = constraint_context.get_range_variable(self.name)
        if ctx_attrs is not None:
            if attrs != ctx_attrs:
                raise VerifyException(
                    f"attributes {tuple(str(x) for x in ctx_attrs)} expected from range variable "
                    f"'{self.name}', but got {tuple(str(x) for x in attrs)}"
                )
        else:
            self.constraint.verify(attrs, constraint_context)
            constraint_context.set_range_variable(self.name, tuple(attrs))

    def get_variable_extractors(
        self,
    ) -> dict[str, VarExtractor[Sequence[AttributeCovT]]]:
        return {self.name: IdExtractor[Sequence[AttributeCovT]]()}

    def can_infer(self, var_constraint_names: Set[str], *, length_known: bool) -> bool:
        return self.name in var_constraint_names

    def infer(
        self, context: ConstraintContext, *, length: int | None
    ) -> Sequence[AttributeCovT]:
        v = context.get_range_variable(self.name)
        return cast(Sequence[AttributeCovT], v)


@dataclass(frozen=True)
class RangeOf(GenericRangeConstraint[AttributeCovT]):
    """
    Constrain each element in a range to satisfy a given constraint.
    """

    constr: GenericAttrConstraint[AttributeCovT]
    _: KW_ONLY
    length: IntConstraint = field(default_factory=AnyInt)

    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        for a in attrs:
            self.constr.verify(a, constraint_context)
        self.length.verify(len(attrs), constraint_context)

    def get_length_extractors(self) -> dict[str, VarExtractor[int]]:
        return self.length.get_length_extractors()

    def can_infer(self, var_constraint_names: Set[str], *, length_known: bool) -> bool:
        return (
            length_known or self.length.can_infer(var_constraint_names)
        ) and self.constr.can_infer(var_constraint_names)

    def infer(
        self,
        context: ConstraintContext,
        *,
        length: int | None,
    ) -> Sequence[AttributeCovT]:
        if length is None:
            length = self.length.infer(context)
        attr = self.constr.infer(context)
        return (attr,) * length


@dataclass(frozen=True)
class SingleOf(GenericRangeConstraint[AttributeCovT]):
    """
    Constrain a range to only contain a single element, which should satisfy a given constraint.
    """

    constr: GenericAttrConstraint[AttributeCovT]

    def verify(
        self,
        attrs: Sequence[Attribute],
        constraint_context: ConstraintContext,
    ) -> None:
        if len(attrs) != 1:
            raise VerifyException(f"Expected a single attribute, got {len(attrs)}")
        self.constr.verify(attrs[0], constraint_context)

    @dataclass(frozen=True)
    class _Extractor(VarExtractor[Sequence[AttributeInvT]]):
        inner: VarExtractor[AttributeInvT]

        def extract_var(self, a: Sequence[AttributeInvT]) -> ConstraintVariableType:
            return self.inner.extract_var(a[0])

    def get_variable_extractors(
        self,
    ) -> dict[str, VarExtractor[Sequence[AttributeCovT]]]:
        return {
            v: self._Extractor(r)
            for v, r in self.constr.get_variable_extractors().items()
        }

    def can_infer(
        self, var_constraint_names: Set[str], *, length_known: int | None
    ) -> bool:
        return self.constr.can_infer(var_constraint_names)

    def infer(
        self, context: ConstraintContext, *, length: int | None
    ) -> Sequence[AttributeCovT]:
        return (self.constr.infer(context),)


def range_constr_coercion(
    attr: (
        AttributeCovT
        | type[AttributeCovT]
        | GenericAttrConstraint[AttributeCovT]
        | GenericRangeConstraint[AttributeCovT]
    ),
) -> GenericRangeConstraint[AttributeCovT]:
    if isinstance(attr, GenericRangeConstraint):
        return attr
    return RangeOf(attr_constr_coercion(attr), length=AnyInt())


def single_range_constr_coercion(
    attr: AttributeCovT | type[AttributeCovT] | GenericAttrConstraint[AttributeCovT],
) -> GenericRangeConstraint[AttributeCovT]:
    return SingleOf(attr_constr_coercion(attr))
